#!/usr/bin/env python

from .find_doppler import FindDoppler

import sys
import os
import logging
import numpy as np
import time
from argparse import ArgumentParser


def main(args=None):
    """This is the entry-point to turboSETI.

    Args:
      args: (Default value = None)

    Returns:

    """
    # Create an option parser to get command-line input/arguments
    p = ArgumentParser(description='turboSETI doppler drift narrowband search utility.')

    p.add_argument('filename', type=str, help='Name of filename to open (h5 or fil)')
    p.add_argument('-M', '--max_drift', dest='max_drift', type=float, default=10.0,
                   help='Set the drift rate to search. Unit: Hz/sec. Default: 10.0')
    p.add_argument('-s', '--snr', dest='snr', type=float, default=25.0,
                   help='SNR threshold. Default: 25.0')
    p.add_argument('-o', '--out_dir', dest='out_dir', type=str, default='./',
                   help='Location for output files. Default: local dir. ')
    p.add_argument('-l', '--loglevel', dest='loglevel', type=str, default='info',
                   help='Specify log level (info, debug)')
    p.add_argument('-c', '--coarse_chans', dest='coarse_chans', type=str, default=None,
                   help='Comma separated list of coarse channels to analyze.')
    p.add_argument('-n', '--n_coarse_chan', dest='n_coarse_chan', type=int, default=None,
                   help='Number of coarse channels in file.')

    if args is None:
        args = p.parse_args()
    else:
        args = p.parse_args(args)

    if args.coarse_chans is None:
        coarse_chans=''
    else:
        coarse_chans = map(int, args.coarse_chans.split(',') )

    # Setting log level
    if args.loglevel == 'info':
        level_log = logging.INFO
    elif args.loglevel == 'debug':
        level_log = logging.DEBUG
    else:
        raise ValueError('Need valid loglevel value (info,debug).')

    if level_log == logging.INFO:
        stream = sys.stdout
        format = '%(name)-15s %(levelname)-8s %(message)s'
    else:
        stream =  sys.stderr
        format = '%%(relativeCreated)5d (name)-15s %(levelname)-8s %(message)s'

    #Doing search
    try:
        t0 = time.time()

        logging.basicConfig(format=format,stream=stream,level = level_log)

        find_seti_event = FindDoppler(args.filename, max_drift=args.max_drift, snr=args.snr, out_dir=args.out_dir,
                                      coarse_chans=coarse_chans, obs_info=None, n_coarse_chan=args.n_coarse_chan)
        find_seti_event.search()

        t1 = time.time()
        print('Search time: %5.2f min' % ((t1-t0)/60.))

    except Exception as e:
        logging.exception(e)
        raise Exception(1,'[turbo_SETI] Some issue with FindDoppler.',e)

if __name__=='__main__':
    main()



import blimpy as bl
from turbo_seti import FindDoppler, seti_event
from turbo_seti import find_event, plot_event
import pylab as plt
import numpy as np
import os
import pytest

HERE = os.path.split(os.path.abspath(__file__))[0]


def find_doppler(filename_fil):
    """ Run turboseti doppler search on filename with default params """
    print("Searching %s" % filename_fil)
    filename_dat = filename_fil.replace('.h5', '.dat')
    filename_log = filename_fil.replace('.h5', 'log')

    if os.path.exists(filename_dat):
        os.remove(filename_dat)

    if os.path.exists(filename_log):
        os.remove(filename_log)

    snr           = 5
    coarse_chans  = ''
    obs_info      = None
    n_coarse_chan = 1
    max_drift     = 1.0

    find_seti_event = FindDoppler(filename_fil, max_drift=max_drift, snr=snr, out_dir=HERE,
                                  coarse_chans=coarse_chans, obs_info=obs_info, n_coarse_chan=n_coarse_chan)
    find_seti_event.search()


def plot_hits(filename_fil, filename_dat):
    """ Plot the hits in a .dat file. """
    table = find_event.read_dat(filename_dat)
    print(table)

    plt.figure(figsize=(10, 8))
    N_hit = len(table)
    if N_hit > 10:
        print("Warning: More than 10 hits found. Only plotting first 10")
        N_hit = 10

    for ii in range(N_hit):
        plt.subplot(N_hit, 1, ii+1)
        plot_event.plot_hit(filename_fil, filename_dat, ii)
    plt.tight_layout()
    plt.savefig(filename_dat.replace('.dat', '.png'))
    plt.show()


def validate_voyager_hits(filename_dat):
    """ This checks voyager hits against known values.

    Known values:
    # --------------------------
    # ID  Drift_Rate SNR 	      Unc_Freq 	      Corr_Freq 	Index   	freq_start 	   freq_end 	...
    # --------------------------
    001	 -0.392226	 30.612128	   8419.319368	   8419.319368	739933	   8419.319779	   8419.318963	...
    002	 -0.373093	245.707984	   8419.297028	   8419.297028	747929	   8419.297439	   8419.296623	...
    003	 -0.392226	 31.220652	   8419.274374	   8419.274374	756037	   8419.274785	   8419.273969	...
    (from flipped)
    003	 -0.392226	 30.612118	   8419.319366	   8419.319366	308642	   8419.318955	   8419.319771	...
    002	 -0.373093	245.707905	   8419.297025	   8419.297025	300646	   8419.296614	   8419.297430	...
    001	 -0.392226	 31.220642	   8419.274372	   8419.274372	292538	   8419.273961	   8419.274777	...
    """
    h = find_event.read_dat(filename_dat)

    valid_data = [
        {
            'Freq': 8419.319368,
            'FreqStart': 8419.319779,
            'FreqEnd': 8419.318963,
            'SNR': 30.612128,
            'DriftRate': -0.392226,
        },
        {
            'Freq': 8419.297028,
            'FreqStart': 8419.297439,
            'FreqEnd': 8419.296623,
            'SNR': 245.707984,
            'DriftRate': -0.373093,
        },
        {
            'Freq': 8419.274374,
            'FreqStart': 8419.274785,
            'FreqEnd': 8419.273969,
            'SNR': 31.220652,
            'DriftRate': -0.392226,
        }
    ]

    atols = {'Freq': 0.000005, 'FreqStart': 0.00001, 'FreqEnd': 0.00001, 'SNR': 0.001, 'DriftRate': 0.02}

    for vd in valid_data:
        hmax = h[np.isclose(h['Freq'], vd['Freq'], rtol=0.000001)].iloc[0]

        for key in vd.keys():
            print(key, hmax[key], vd[key])
            if key in ('FreqStart', 'FreqEnd'):
                upper = np.isclose(hmax[key], vd['FreqStart'], atol=atols[key], rtol=0)
                lower = np.isclose(hmax[key], vd['FreqEnd'], atol=atols[key], rtol=0)
                assert upper or lower
            else:
                assert np.isclose(hmax[key], vd[key], atol=atols[key], rtol=0)
    return hmax


def test_find_doppler_voyager():
    """ Run turboseti on Voyager data """
    filename_fil = os.path.join(HERE, 'Voyager1.single_coarse.fine_res.h5')
    filename_dat = filename_fil.replace('.h5', '.dat')
    find_doppler(filename_fil)
    validate_voyager_hits(filename_dat)
    plot_hits(filename_fil, filename_dat)


def test_find_doppler_voyager_flipped():
    """ Run turboseti on Voyager data (flipped in frequency) """
    filename_fil = os.path.join(HERE, 'Voyager1.single_coarse.fine_res.flipped.h5')
    filename_dat = filename_fil.replace('.h5', '.dat')
    find_doppler(filename_fil)
    validate_voyager_hits(filename_dat)
    plot_hits(filename_fil, filename_dat)

def test_find_doppler_voyager_filterbank():
    """ Run turboseti on Voyager data (filterbank version) """
    filename_fil = os.path.join(HERE, 'Voyager1.single_coarse.fine_res.fil')
    filename_dat = filename_fil.replace('.fil', '.dat')
    find_doppler(filename_fil)
    #validate_voyager_hits(filename_dat)
    #plot_hits(filename_fil, filename_dat)


def test_turboSETI_entry_point():
    """ Test the command line utility turboSETI """

    filename_fil = os.path.join(HERE, 'Voyager1.single_coarse.fine_res.flipped.h5')
    args = [filename_fil, ]
    seti_event.main(args)

def test_plotting():
    """ Some basic plotting tests

    TODO: Improve these tests (and the functions for that matter!
    """
    filename_fil = os.path.join(HERE, 'Voyager1.single_coarse.fine_res.h5')
    fil = bl.Waterfall(filename_fil)

    # Test make_waterfall_plots -- needs 6x files
    filenames_list = [filename_fil] * 6
    target  = 'Voyager'
    drates  = [-0.392226]
    fvals   = [8419.274785]
    f_start = 8419.274374 - 600e-6
    f_stop  = 8419.274374 + 600e-6
    node_string = 'test'
    filter_level = 1
    plot_event.make_waterfall_plots(filenames_list, target, drates, fvals, f_start, f_stop, node_string, filter_level)
    plt.show()

def test_data_handler():
    """ Basic data handler test """
    from turbo_seti.find_doppler import data_handler
    with pytest.raises(AttributeError):
        fh = data_handler.DATAHandle(filename='made_up_not_existing_file.h5')

if __name__ == "__main__":

    #test_turboSETI_entry_point()
    #test_find_doppler_voyager()
    #test_find_doppler_voyager_flipped()
    #test_plotting()
    #test_find_doppler_voyager_filterbank()
    test_data_handler()



#!/usr/bin/env python
"""
Script to plot events
    ..author: Sofia Sheikh (ssheikhmsa@gmail.com)
    ..modified to work for GC data by Karen Perez (kip8@cornell.edu)
"""

#General packages import
import glob
import os
import sys
import pandas as pd
#import pylab as plt
import numpy as np
import logging; logging.disable(logging.CRITICAL);

#BL + my packages import
#import updated_find_event
import pandas as pd
import blimpy as bl
from astropy.time import Time

#Plotting packages import
import matplotlib
import matplotlib.pyplot as plt

from .find_event import make_table

#preliminary plot arguments
fontsize=16
font = {'family' : 'DejaVu Sans',
'size' : fontsize}
MAX_IMSHOW_POINTS = (10096, 10096)


def plot_hit(fil_filename, dat_filename, hit_id, bw=None, offset=0):
    """Plot a candidate from a .dat file

    Args:
      fil_filename(str): Path to filterbank file to plot
      dat_filename(str): Path to turbosSETI generated .dat output file of events
      hit_id(int): ID of hit in the dat file to plot (TopHitNum)
      offset(float, optional): Offset drift line on plot. Default 0.
      bw:  (Default value = None)

    Returns:

    """
    # Load hit details
    dat = make_table(dat_filename)
    hit = dat.iloc[hit_id]

    f0 = hit['Freq']

    if bw is None:
        bw_mhz = np.abs(hit['FreqStart'] - hit['FreqEnd'])
    else:
        bw_mhz = bw * 1e-6

    fil = bl.Waterfall(fil_filename, f_start=f0 - bw_mhz / 2, f_stop=f0 + bw_mhz / 2)
    t_duration = (fil.n_ints_in_file - 1) * fil.header['tsamp']

    fil.plot_waterfall()
    overlay_drift(f0, hit['DriftRate'], t_duration, offset)


def rebin(d, n_x, n_y):
    """Rebin data by averaging bins together

    Args:
      d(np.array): data
      n_x(int): number of bins in x dir to rebin into one
      n_y: 

    Returns:
      d: rebinned data with shape (n_x, n_y)

    """

    if d.ndim == 2:
        if n_y is None:
            n_y = 1
        if n_x is None:
            n_x = 1
        print ('d.shape[0]', d.shape[0])
        print ('d.shape[1]', d.shape[1])
        print ('n_x', n_x)
        print ('n_y', n_y)
        d = d[:int(d.shape[0] // n_x) * n_x, :int(d.shape[1] // n_y) * n_y]
        d = d.reshape((d.shape[0] // n_x, n_x, d.shape[1] // n_y, n_y))
        d = d.mean(axis=3)
        d = d.mean(axis=1)
    elif d.ndim == 1:
        d = d[:int(d.shape[0] // n_x) * n_x]
        d = d.reshape((d.shape[0] // n_x, n_x))
        d = d.mean(axis=1)
    else:
        raise RuntimeError("Only NDIM <= 2 supported")
    return d


def overlay_drift(f_event, drate, t_duration, offset=0):
    """Overlay drift line on plot

    Args:
      f_event(float): Start frequency of event in MHz
      drate(float): Drift rate in Hz/s
      t_duration(float): Time duration in seconds
      offset(float, optional): Offset line from event by amount. Default 0.

    Returns:

    """
    if offset == 'auto':
        offset = - 0.2 * drate*t_duration
    plt.plot((f_event+offset/1e6, f_event+drate/1e6*t_duration+offset/1e6), (0, t_duration), c='#cc0000', ls='dashed', lw=2)


def make_waterfall_plots(filenames_list, target, drates, fvals, f_start,f_stop, node_string, filter_level, ion=False,
                         epoch=None,bw=250.0, local_host='',plot_name='',save_pdf_plot=False,saving_fig=False,offset=0,
                         dedoppler=False,**kwargs):
    """Makes waterfall plots per group of ON-OFF pairs (up to 6 plots.)

    Args:
      filenames_list: 
      target: 
      drates: 
      fvals: 
      f_start: 
      f_stop: 
      node_string: 
      filter_level: 
      ion:  (Default value = False)
      epoch:  (Default value = None)
      bw:  (Default value = 250.0)
      local_host:  (Default value = '')
      plot_name:  (Default value = '')
      save_pdf_plot:  (Default value = False)
      saving_fig:  (Default value = False)
      offset:  (Default value = 0)
      dedoppler:  (Default value = False)
      **kwargs: 

    Returns:

    """
    
    #prepares for plotting
    print('Preparing to plot: ', target)
    matplotlib.rc('font', **font)
    if ion:
        plt.ion()

    #defines a minimum and maximum of... something
    min_val = 0
    max_val = 5.
    factor = 1e6
    units = 'Hz'

    #sets up the sub-plots
    n_plots = len(filenames_list)
    fig = plt.subplots(n_plots, sharex=True, sharey=True,figsize=(10, 2*n_plots))

    #finding plotting values range for the first panel (A1)
    fil = bl.Waterfall(filenames_list[0], f_start=f_start, f_stop=f_stop)
    
    filheader=fil.header
    f={}
    for key, value in filheader.items():
        f[key]=value
    filheader=f

    print ('filheader', filheader)
    t0 = filheader['tstart']
    plot_f, plot_data = fil.grab_data(f_start=f_start, f_stop=f_stop)
    dec_fac_x, dec_fac_y = 1, 1
    #print ('plot_data', plot_data)

    #rebinning data to plot correctly with fewer plots
    if plot_data.shape[0] > MAX_IMSHOW_POINTS[0]:
        dec_fac_x = plot_data.shape[0] / MAX_IMSHOW_POINTS[0] 
        print ('dec_fac_x', dec_fac_x)
    if plot_data.shape[1] > MAX_IMSHOW_POINTS[1]:
        dec_fac_y =  plot_data.shape[1] /  MAX_IMSHOW_POINTS[1]
        print ('dec_fac_y', dec_fac_y)
    #insert rebin definition code from utils.py in blimpy and edit it because this is the problem
    d=plot_data
    n_x=dec_fac_x
    n_y=dec_fac_y

    d = rebin(d, n_x, n_y)
    plot_data=d
    #print('d', d)

    #plot_data = rebin(plot_data, dec_fac_x, dec_fac_y)

    #investigate intensity values for A1 (first panel)
    plot_data = 10*np.log10(plot_data)
    A1_avg = np.median(plot_data)
    A1_max = plot_data.max()
    A1_std = np.std(plot_data)
    
    #defining more plot parameters
    delta_f = 0.000250
    epoch = filheader['tstart']
    mid_f = np.abs(f_start+f_stop)/2.
    drate_max = np.max(np.abs(drates))
    
    subplots = []
    
    #working out intensity scale
    if kwargs.get('clim', None) is None:
        vmin=A1_avg-A1_std*min_val-2
        vmax=A1_avg+max_val*A1_std
    else:
        vmin, vmax = kwargs['clim']
        
    #Filling in each subplot for the full plot
    for i,filename in enumerate(filenames_list):
        subplot = plt.subplot(n_plots,1,i+1)
        subplots.append(subplot)
        fil = bl.Waterfall(filename, f_start=f_start, f_stop=f_stop)
        filheader=fil.header


        try:
            this_plot = plot_waterfall(fil,f_start=f_start, f_stop=f_stop, drate=drate_max,
                                       vmin=vmin,vmax=vmax,**kwargs)
            for drate, fval in zip(drates, fvals):
                t_elapsed = Time(filheader['tstart'], format='mjd').unix - Time(t0, format='mjd').unix
                t_duration = (fil.n_ints_in_file -1)* filheader['tsamp']
                f_event = fval + drate / 1e6 * t_elapsed
                overlay_drift(f_event, drate, t_duration, offset)
        except:
            raise
            
    #Titling the plot
        if i == 0:
            srcname = "%s $\dot{\\nu}$=%2.3f Hzs$^{-1}$" % (target, drate_max)
            plt.title(srcname)
    #Plot formatting
        if i < len(filenames_list)-1:
            plt.xticks(np.arange(f_start, f_stop, delta_f/4.), ['','','',''])

    #More plot formatting.
    ax = plt.gca()
    ax.get_xaxis().get_major_formatter().set_useOffset(False)
    xloc = np.linspace(f_start, f_stop, 5)
    xticks = [round(loc_freq) for loc_freq in (xloc - mid_f)*factor]
    if np.max(xticks) > 1000:
        xticks = [xt/1000 for xt in xticks]
        units = 'kHz'
    plt.xticks(xloc, xticks)
    plt.xlabel("Relative Frequency [%s] from %f MHz"%(units,mid_f),fontdict=font)

    #Colorbar
    cax = fig[0].add_axes([0.94, 0.11, 0.03, 0.77])
    fig[0].colorbar(this_plot,cax=cax,label='Power [dB counts]')
    
    #Adjust plots
    plt.subplots_adjust(hspace=0,wspace=0)
    
    #save the figures
    plt.savefig(node_string + '_f' + str(filter_level) + '_' + target[0] + '_dr_' + "{:0.2f}".format(drate_max) + '_freq_' "{:0.2f}".format(f_start) + ".png",
               bbox_inches='tight')

    return subplots


def plot_waterfall(fil, f_start=None, f_stop=None, drate=None, if_id=0, logged=True, cb=False,freq_label=False,MJD_time=False, **kwargs):
    """Plot waterfall of data

    Args:
      f_start(float, optional): start frequency, in MHz (Default value = None)
      f_stop(float, optional): stop frequency, in MHz (Default value = None)
      logged(bool, optional): Plot in linear (False) or dB units (True), (Default value = True)
      cb(bool, optional): for plotting the colorbar (Default value = False)
      kwargs: keyword args to be passed to matplotlib imshow()
      fil: 
      drate:  (Default value = None)
      if_id:  (Default value = 0)
      freq_label:  (Default value = False)
      MJD_time:  (Default value = False)
      **kwargs: 

    Returns:

    """

    #prepare font
    matplotlib.rc('font', **font)

    #Get the data
    plot_f, plot_data = fil.grab_data(f_start=f_start, f_stop=f_stop)

    # Make sure waterfall plot is under 4k*4k
    dec_fac_x, dec_fac_y = 1, 1
    if plot_data.shape[0] > MAX_IMSHOW_POINTS[0]:
        dec_fac_x = plot_data.shape[0] / MAX_IMSHOW_POINTS[0]
    if plot_data.shape[1] > MAX_IMSHOW_POINTS[1]:
        dec_fac_y =  plot_data.shape[1] /  MAX_IMSHOW_POINTS[1]
    #insert rebin definition code from utils.py in blimpy and edit it because this is the problem
    d=plot_data
    n_x=dec_fac_x
    n_y=dec_fac_y

    d = rebin(d, n_x, n_y)
    plot_data=d
    #print('d', d)

    #plot_data = rebin(plot_data, dec_fac_x, dec_fac_y)

    if MJD_time:
        extent=(plot_f[0], plot_f[-1], fil.timestamps[-1], fil.timestamps[0])
    else:
        extent=(plot_f[0], plot_f[-1], (fil.timestamps[-1]-fil.timestamps[0])*24.*60.*60, 0.0)

    #plots and scales intensity
    kwargs['cmap'] = kwargs.get('cmap', 'viridis')
    kwargs['logged'] = True
    if kwargs['logged'] == True:
        plot_data = 10*np.log10(plot_data)
        kwargs.pop('logged')

    #shows the waterfall plot
    this_plot = plt.imshow(plot_data,
        aspect='auto',
        rasterized=True,
        interpolation='nearest',
        extent=extent,
        **kwargs
    )
    #add colorbar
    if cb:
        plt.colorbar()

    #add plot labels
    if freq_label:
        plt.xlabel("Frequency [Hz]",fontdict=font)
    if MJD_time:
        plt.ylabel("Time [MJD]",fontdict=font)
    else:
        plt.ylabel("Time [s]",fontdict=font)

    return this_plot 

def plot_candidate_events_individually(full_candidate_event_dataframe, correct_fils, source_name,
                                       node_string, filter_level, show=False, overwrite=False, offset=0, **kwargs):
    """

    Args:
      full_candidate_event_dataframe: 
      correct_fils: 
      source_name: 
      node_string: 
      filter_level: 
      show:  (Default value = False)
      overwrite:  (Default value = False)
      offset:  (Default value = 0)
      **kwargs: 

    Returns:

    """
    trimmed_candidate_event_dataframe=pd.read_csv(full_candidate_event_dataframe)
    print ('full_candidate_event_dataframe', full_candidate_event_dataframe)
    #get only the events in the dataframe that are from the right target
    #full_candidate_event_dataframe=full_candidate_event_dataframe['Source']
    print ('trimmed_candidate_event_dataframe_Source', trimmed_candidate_event_dataframe.loc[:, ['Source']])
    #print ('full_candidate_event_dataframe_Source', full_candidate_event_dataframe['Source'])
    candidate_event_dataframe = trimmed_candidate_event_dataframe.loc[:, ['Source', 'Freq', 'DriftRate','FreqStart']]
    #candidate_event_dataframe = full_candidate_event_dataframe.loc[full_candidate_event_dataframe['Source'] == source_name]
    print ('full_candidate_event_dataframe', candidate_event_dataframe)

    #load in the data for each individual hit
    for i in range(0, len(candidate_event_dataframe)):
        candidate = candidate_event_dataframe.iloc[i]
        source_id = candidate['Source']
        f_mid = candidate['Freq']
        drate = -1*candidate['DriftRate']
                
        #load in the list of .fil files
        filelist = open(correct_fils).readlines()
        filelist = [files.replace('\n','') for files in filelist]
        #print ('dat_file_list', dat_file_list)
        filelist = [files.replace(',','') for files in filelist]
        #filelist = correct_fils    
        print ('filelist', filelist) 
        #calculate the length of the total ABABAB from the fil files' headers
        fil0 = bl.Waterfall(filelist[0], load_data=True)
        print ('fil0', fil0)
        fil0header=fil0.header
        print ('fil0header', fil0header)
        y={}
        for key, value in fil0header.items():
            y[key.decode("utf-8")]=value
        fil0header=y
        print ('fil0header', fil0header)
        t0 = fil0header['tstart']
        
        fil6 = bl.Waterfall(filelist[-1], load_data=True)
        fil6header=fil6.header
        z={}
        for key, value in fil6header.items():
            z[key.decode("utf-8")]=value
        fil6header=z
        t6 = fil6header['tstart']
        t_elapsed = Time(t6, format='mjd').unix - Time(t0, format='mjd').unix + (fil6.n_ints_in_file -1) * fil6header['tsamp']


        #calculate the width of the plot based on making sure the full drift is visible
        bw = 2.4*abs(drate)/1e6 * t_elapsed
        bw = np.max((bw, 500./1e6))

        #Get start and stop frequencies based on midpoint and bandwidth
        f_start, f_stop = np.sort((f_mid-bw/2,  f_mid+bw/2))

        #Print useful values
        print('*************************************************')
        print('***     The Parameters for This Plot Are:     ***')
        print('*************************************************')   
        print('Target = ', source_id)
        print('Bandwidth (MHz) = ', bw)
        print('Total Time Elapsed (s) = ', t_elapsed)
        print('Start Frequency (MHz) = ', f_start)
        print('Middle Frequency (MHz) = ', f_mid)
        print('Stop Frequency (MHz) = ', f_stop)
        print('Expected Drift (Hz/s) = ', drate)
        print('*************************************************')
        print('*************************************************')
                
        #Pass info to make_waterfall_plots() function
        subplots = make_waterfall_plots(filelist, 
                                        [source_id], 
                                        [drate], 
                                        [f_mid], 
                                        f_start, 
                                        f_stop, 
                                        node_string,
                                        filter_level,
                                        ion=False, 
                                        epoch=None,
                                        local_host='',
                                        plot_name='',
                                        save_pdf_plot=True,
                                        saving_fig=True, 
                                        offset=offset, 
                                        **kwargs)
    return 



#!/usr/bin/env python
"""
Backend script to find drifting, narrowband events in a generalized cadence of 
ON-OFF radio SETI observations.

In this code, the following terminology is used:
Hit = single strong narrowband signal in an observation
Event = a strong narrowband signal that is associated with multiple hits
        across ON observations
    
The main function contained in this file is *find_events*
    Find_events uses the other helper functions in this file (described below)
    to read a list of turboSETI .dat files. It then finds events within this 
    group of files. 
    
The following helper functions are contained in this file:
    end_search      - prints the runtime and ends the program
    make_table      - generates a pandas dataframe from a turboSETI
                      output .dat file created by seti_event.py
    calc_freq_range - Calculates a range of frequencies where RFI in an OFF 
                      could be related to a hit in an ON given a freq and 
                      drift_rate
    follow_event    - Follows a given hit to the next observation of the ON
                      target and looks for hits which could be part of the 
                      same event
                      
Usage (beta):
    It is highly recommended that users interact with this program via the
    front-facing find_event_pipeline.py script. See the usage of that file in
    its own documentation. 
    
    If you would like to run find_events without calling
    find_event_pipeline.py, the usage is as follows:
    
    find_event.find_events(file_sublist, 
                           SNR_cut=10, 
                           check_zero_drift=False, 
                           filter_threshold=3, 
                           on_off_first='ON')
    
    file_sublist        A Python list of .dat files with ON observations of a
                        single target alternating with OFF observations. This 
                        cadence can be of any length, given that the ON source 
                        is every other file. This includes Breakthrough Listen 
                        standard ABACAD as well as OFF first cadences like 
                        BACADA. Minimum cadence length is 2, maximum cadence 
                        length is unspecified (currently tested up to 6).
                   
    SNR_cut             The threshold SNR below which hits in the ON source 
                        will be disregarded. For the least strict thresholding, 
                        set this parameter equal to the minimum-searched SNR 
                        that you used to create the .dat files from 
                        seti_event.py. Recommendation (and default) is 10.
                   
    check_zero_drift    A True/False flag that tells the program whether to
                        include hits that have a drift rate of 0 Hz/s. Earth-
                        based RFI tends to have no drift rate, while signals
                        from the sky are expected to have non-zero drift rates.
                        Default is False.
                        
    filter_threshold    Specification for how strict the hit filtering will be.
                        There are 3 different levels of filtering, specified by
                        the integers 1, 2, and 3. Filter_threshold = 1 
                        returns hits above an SNR cut, taking into account the
                        check_zero_drift parameter, but without an ON-OFF check.
                        Filter_threshold = 2 returns hits that passed level 1
                        AND that are in at least one ON but no OFFs. 
                        Filter_threshold = 3 returns events that passed level 2
                        AND that are present in *ALL* ONs. Default is 3.
                        
    on_off_first        Tells the code whether the .dat sequence starts with
                        the ON or the OFF observation. Valid entries are 'ON'
                        and 'OFF' only. Default is 'ON'.
                    
author: 
    Version 2.0 - Sofia Sheikh (ssheikhmsa@gmail.com)
    Version 1.0 - Emilio Enriquez (jeenriquez@gmail.com)
    
Last updated: 04/08/2020

***
NOTE: This code works for .dat files that were produced by seti_event.py
after turboSETI version 0.8.2, and blimpy version 1.1.7 (~mid 2019). The 
drift rates *before* that version were recorded with the incorrect sign
and thus the drift rate sign would need to be flipped in the make_table 
function.
***

"""

import pandas as pd
import numpy as np
import time

pd.options.mode.chained_assignment = None  
#^To remove pandas warnings: default='warn'

#------
#Hardcoded values - SZS NOTE: Make these NOT Hardcoded
MAX_DRIFT_RATE = 2.0    # NOTE: these two values need to be updated.
OBS_LENGTH = 300.
#------


def end_search(t0):
    """ends the search when there are no candidates left, or when the filter
    level matches the user-specified level

    Args:
      t0: 

    Returns:

    """
    #Report elapsed search time
    t1 = time.time()
    print('Search time: %.2f sec' % ((t1-t0)))
    print('------   o   -------')
    return

def read_dat(filename):
    """Read a turboseti .dat file

    Args:
      filename(str): Name of .dat file to open

    Returns:
        pandas dataframe of hits
    """

    file_dat = open(filename.strip())
    hits = file_dat.readlines()

    # Get info from the .dat file header
    FileID = hits[1].strip().split(':')[-1].strip()
    Source = hits[3].strip().split(':')[-1].strip()

    MJD = hits[4].strip().split('\t')[0].split(':')[-1].strip()
    RA = hits[4].strip().split('\t')[1].split(':')[-1].strip()
    DEC = hits[4].strip().split('\t')[2].split(':')[-1].strip()

    DELTAT = hits[5].strip().split('\t')[0].split(':')[-1].strip()  # s
    DELTAF = hits[5].strip().split('\t')[1].split(':')[-1].strip()  # Hz

    # Get info from individual hits (the body of the .dat file)
    all_hits = [hit.strip().split('\t') for hit in hits[9:]]

    # Now reorganize that info to be grouped by column (parameter)
    # not row (individual hit)
    if all_hits:
        TopHitNum = list(zip(*all_hits))[0]
        DriftRate = [float(df) for df in list(zip(*all_hits))[1]]
        SNR = [float(ss) for ss in list(zip(*all_hits))[2]]
        Freq = [float(ff) for ff in list(zip(*all_hits))[3]]
        ChanIndx = list(zip(*all_hits))[5]
        FreqStart = list(zip(*all_hits))[6]
        FreqEnd = list(zip(*all_hits))[7]
        CoarseChanNum = list(zip(*all_hits))[10]
        FullNumHitsInRange = list(zip(*all_hits))[11]

        data = {'TopHitNum': TopHitNum,
                'DriftRate': DriftRate,
                'SNR': SNR,
                'Freq': Freq,
                'ChanIndx': ChanIndx,
                'FreqStart': FreqStart,
                'FreqEnd': FreqEnd,
                'CoarseChanNum': CoarseChanNum,
                'FullNumHitsInRange': FullNumHitsInRange
                }

        # Creating pandas dataframe from data we just read in
        df_data = pd.DataFrame(data)
        df_data = df_data.apply(pd.to_numeric)

    else:
        df_data = pd.DataFrame()

    # Matching column information from before to the .dat data we read in
    df_data['FileID'] = FileID
    df_data['Source'] = Source.upper()
    df_data['MJD'] = MJD
    df_data['RA'] = RA
    df_data['DEC'] = DEC
    df_data['DELTAT'] = DELTAT
    df_data['DELTAF'] = DELTAF

    # Adding extra columns that will be filled out by this program
    df_data['Hit_ID'] = ''
    df_data['status'] = ''
    df_data['in_n_ons'] = ''
    df_data['RFI_in_range'] = ''

    return df_data

def make_table(filename, init=False):
    """Creates a pandas dataframe with column names standard for turboSETI .dat
    output files, either directly (if) or by reading the file line-by line and
    then reorganizing the output (else)

    Args:
      filename: 
      init:  (Default value = False)

    Returns:
        df_data: pandas dataframe

    """
    
    if init:
        columns = ['FileID','Source','MJD','RA','DEC', 'DELTAT','DELTAF',
                   'TopHitNum','DriftRate', 'SNR', 'Freq', 'ChanIndx', 
                   'FreqStart', 'FreqEnd', 'CoarseChanNum', 
                   'FullNumHitsInRange','status','Hit_ID','in_n_ons',
                   'RFI_in_range']
        df_data = pd.DataFrame(columns=columns)

    else:
        df_data = read_dat(filename)
    return df_data

def calc_freq_range(hit,delta_t=0,max_dr=True,follow=False):
    """Calculates a range of frequencies where RFI in an off-source could
        be related to a hit in an on-source given a freq and drift_rate.

    Args:
      hit: 
      delta_t:  (Default value = 0)
      max_dr:  (Default value = True)
      follow:  (Default value = False)

    Returns:
        list [low_bound, high_bound]
    """
    if max_dr:
        drift_rate = MAX_DRIFT_RATE
    else:
        drift_rate = hit['DriftRate']
        
    if drift_rate == 0.0:
        drift_rate = float(hit['DELTAF'])/float(hit['DELTAT'])
    if follow:
        freq = hit['Freq'] + drift_rate*(delta_t)/1e6
        delta_t = 2*OBS_LENGTH  # NOTE: I'm doing twice the amount. To widen the range a bit. 
                                # Still it is dependent on hit['DriftRate']
    else:
        freq = hit['Freq']
        delta_t = delta_t+OBS_LENGTH  #adding to the end of the obs.
    
    low_bound  = freq - abs(drift_rate)*delta_t/1e6
    high_bound = freq + abs(drift_rate)*delta_t/1e6

    return [low_bound,high_bound]

def follow_event(hit,on_table,get_count=True):
    """Follows a given hit to the next observation of the same target and
    looks for hits which could be part of the same event.

    Args:
      hit: 
      on_table: 
      get_count:  (Default value = True)

    Returns:
        new_on_table or count

    """

    #uses calc_freq_range to see how much the hit *should* have drifted by
    freq_range = calc_freq_range(hit,delta_t=on_table['delta_t'].values[0],max_dr=False,follow=True)
    
    #looks at the on (next given observation) to see if there are any
    #hits that could plausibly be related to the first one
    new_on_table = on_table[(on_table['Freq']>freq_range[0]) & (on_table['Freq']<freq_range[1])]

    #we only care if there were or were not plausible hits, so turn output into
    #a binary variable with get_count=True
    if get_count:
        n_hits_in_range = len(new_on_table)

        #Avoiding cases where multiple hits in one obs, and none in the other.
        if n_hits_in_range:
            return 1
        else:
            return 0
    else:
        return new_on_table

def find_events(dat_file_list,  SNR_cut=10, check_zero_drift=False, filter_threshold=3, on_off_first='ON'):
    """Reads a list of turboSETI .dat files.
        It calls other functions to find events within this group of files.
        Filter_threshold allows the return of a table of events with hits at
        different levels of filtering.
        Filter_threshold = [1,2,3] means:
            1) Hits above an SNR cut witout AB check
            2) Hits that are only in some As and no Bs
            3) Hits that are only in all As and no Bs

    Args:
      dat_file_list: 
      SNR_cut:  (Default value = 10)
      check_zero_drift:  (Default value = False)
      filter_threshold:  (Default value = 3)
      on_off_first:  (Default value = 'ON')

    Returns:

    """
    #Initializing timer
    t0 = time.time()
    
    print('------   o   -------')
    print("Loading data...")
    
    #Preparing to read in the list of files
    on_table_list = []
    off_table_list = []
    off_count = 1 
    on_count = 1  
    
    on_off_indicator = 0
    number_of_ons = int(np.floor(len(dat_file_list) / 2.0))
    if on_off_first == 'ON':
        on_off_indicator = 1
        number_of_ons = int(np.ceil(len(dat_file_list) / 2.0))
    
    for i,dat_file in enumerate(dat_file_list):
        #Checking if the file is an on or off observation, based on index
        if i%2 == on_off_indicator:
            #Using make_table function to read the .dat file 
            #and create the pandas hit table
            off_table_i=make_table(dat_file)
            off_table_i['status'] = 'off_table_%i'%off_count
            print('Loaded %i hits from %s'%(len(off_table_i), dat_file))

            #Grouping all of the off hits into one table
            off_table_list.append(off_table_i)
            off_count+=1
            
        else: 
            #Using make_table function to read the .dat file 
            #and create the pandas hit table
            on_table_i=make_table(dat_file)
            on_table_i['status'] = 'on_table_%i'%on_count
            print('Loaded %i hits from %s'%(len(on_table_i), dat_file))
    
            #Grouping all of the on hits into one table
            on_table_list.append(on_table_i)
            off_count+=1
    
    #If there are no hits on any on target, end the program
    if not len(on_table_list):
        print('There are no hits in this cadence :(')
        end_search(t0)
        return
    
    #Concatenating the on and off tables into a giant on table 
    #and a giant off table
    on_table = pd.concat(on_table_list,ignore_index=True)
    off_table = pd.concat(off_table_list,ignore_index=True)        
    
    #Check that all targets in the on_table come from the same source
    if on_table['Source'].unique().shape[0] > 1:
        raise ValueError('There are multiple sources in the on table.' 
                         'Please check your input files, ' 
                         'on_off_first parameter.')
    
    #Obtain the start times for each hit in the first on table
    ref_time = float(on_table[on_table['status'] == 'on_table_1']['MJD'].unique()[0])
    #Calculating and saving delta_t, in seconds, to follow a given hit from 
    #the first on table to see if it appears in the following on tables
    on_table['delta_t'] = on_table['MJD'].apply(lambda x: 
        (float(x) - ref_time)*3600*24)

    #######################################################################
    print('All data loaded!')
    print()
    print('Finding events in this cadence...')
    #######################################################################
    #Using logic statements and pandas capabilities to find events that:
    # 1) Are above an SNR cut, taking into account the check_zero_drift parameter, 
    #    but without an ON-OFF check.
    # 2) Passed level 1 AND that are in at least one ON but no OFFs 
    # 3) Passed level 2 AND that are present in *ALL* ONs.

    #Optionally remove signals that don't have a drift rate
    if check_zero_drift:
        zero_adjusted_table = on_table
    else:
        zero_adjusted_table = on_table[on_table['DriftRate'] != 0.0]

    #Remove signals below a certain signal-to-noise ratio (SNR_cut)
    snr_adjusted_table = zero_adjusted_table[zero_adjusted_table['SNR'] > SNR_cut]
    
    #If there are no hits left once SNR adjustments are made, 
    #let the user know, return the zero-drift adjusted table, and end the
    #search
    if (len(snr_adjusted_table) == 0):
        print('Found no hits above the SNR cut :(')
        end_search(t0)
        return
    if filter_threshold == 1:
        print('Found a total of %i hits above the SNR cut in this cadence!'%len(snr_adjusted_table))
        print('Filter level is 1 - returning this table...')
        end_search(t0)
        return snr_adjusted_table
    else:
        print('Found a total of %i hits above the SNR cut in this cadence!'%len(snr_adjusted_table))
    
    #----------------------------------------------------------------------

    #Now find how much RFI is within a frequency range of the hit 
    #by comparing the ON to the OFF observations. Update RFI_in_range
    snr_adjusted_table['RFI_in_range'] = snr_adjusted_table.apply(lambda hit: len(off_table[((off_table['Freq'] > calc_freq_range(hit)[0]) & (off_table['Freq'] < calc_freq_range(hit)[1]))]),axis=1)
        
    #If there is no RFI in range of the hit, it graduates to the 
    #not_in_B_table
    not_in_off_table = snr_adjusted_table[snr_adjusted_table['RFI_in_range'] == 0]

    if (len(not_in_off_table) == 0):
        print('Found no hits present in only the on observations in this cadence :(')
        end_search(t0)
        return
    if filter_threshold == 2:    
        print('Found a total of %i hits in only the on observations in this cadence!'%len(not_in_off_table))
        print('Filter level is 2 - returning this table...')
        end_search(t0)
        return not_in_off_table
    else:
        print('Found a total of %i hits in only the on observations in this cadence!'%len(not_in_off_table))
        
    #----------------------------------------------------------------------
    
    on_but_not_off_table_list = []
    in_all_ons_table = []
    #Follow the drifting of a hit to find the events that 
    #are present in ALL on observations
    
    for i in range(1, number_of_ons + 1):
        on_but_not_off_table_list.append(not_in_off_table[not_in_off_table['status'] == 'on_table_%i'%i])
    empty_counter = 0
    for hit_list in on_but_not_off_table_list:
        if hit_list.empty == True:
            empty_counter += 1
    if empty_counter == 0:
        first_on = on_but_not_off_table_list[0]#
        def hit_func(hit):
            """

            Args:
              hit: 

            Returns:

            """
            val = 0
            for i in range(1, len(on_but_not_off_table_list)):
                val += follow_event(hit, on_but_not_off_table_list[i])
            return val
        
        first_on['in_n_ons'] = first_on.apply(hit_func, axis=1)
        in_all_ons_table = first_on[first_on['in_n_ons'] == number_of_ons - 1]
        
        #Create list of events.
        filter_3_event_list = []

        for hit_index, hit in in_all_ons_table.iterrows():
            for table in on_but_not_off_table_list:
                temp_table = follow_event(hit,table,get_count=False)
                temp_table['Hit_ID'] = hit['Source']+'_'+str(hit_index)
                filter_3_event_list += [temp_table]

    else:
        print('NOTE: At least one of the on tables is empty - no events across this cadence :(')
        end_search(t0)
        return

    if len(in_all_ons_table) > 0:
        best_events_table = pd.concat(filter_3_event_list)
        print('Found a total of %i events across this cadence!'%(int(len(best_events_table)/3)))
        end_search(t0)
        return best_events_table
    
    else:
        print('NOTE: Found no events across this cadence :(')
        end_search(t0)
        return
    
    #----------------------------------------------------------------------




#!/usr/bin/env python
"""This is a function to Taylor-tree-sum a data stream. It assumes that
the arrangement of data stream is, all points in first spectra, all
points in second spectra, etc...  Data are summed across time.
         Original algorithm: R. Ramachandran, 07-Nov-97, nfra.
         Modified 2011 A. Siemion float/64 bit addressing (C-code)
         Modified 2014 H. Chen python version
         Modified 1-Feb-2016 E. Enriquez + P.Schellart cython version
"""

import numpy as np
cimport numpy as np

cimport cython

DTYPE = np.float64
# "ctypedef" assigns a corresponding compile-time type to DTYPE_t. For
# every type in the numpy module there's a corresponding compile-time
# type with a _t-suffix.
ctypedef np.float64_t DTYPE_t

@cython.boundscheck(False)
def taylor_flt(np.ndarray[DTYPE_t, ndim=1] outbuf, long mlen, long nchn):
    """
    Parameters:
        outbuf       : input array (float), replaced by dedispersed data
                       at the output
        mlen         : dimension of outbuf[] (long int)
        nchn         : number of frequency channels (long int)
    """

    assert outbuf.dtype == DTYPE

    cdef long nsamp = (mlen/nchn) - (2*nchn)
    cdef long npts = nsamp + nchn
    cdef int nstages = int(np.log2(nchn))

    cdef int istages
    cdef int isec
    cdef int ipair
    cdef int i1
    cdef int ndelay
    cdef int ndelay2
    cdef int nmem = 1
    cdef int nmem2
    cdef int nsec1

    cdef int ndat1 = nsamp + 2 * nchn
    cdef int koff
    cdef int nfin
    cdef int ioff1
    cdef int i2

    cdef DTYPE_t itemp

    for istages in range(0, nstages):
        nmem  *= 2
        nsec1  = nchn/nmem
        nmem2  = nmem - 2
        for isec in range(0, nsec1):
            ndelay = -1
            koff = isec * nmem
            for ipair in range(0, nmem2+1, 2):
                ioff1 = (bitrev(ipair, istages+1) + koff) * ndat1
                i2 = (bitrev(ipair+1, istages+1) + koff) * ndat1
                ndelay += 1
                ndelay2 = (ndelay + 1)
                nfin = (npts + ioff1)

                for i1 in range(ioff1, nfin):
                    itemp = outbuf[i1] + outbuf[i2+ndelay]
                    outbuf[i2] = outbuf[i1] + outbuf[i2+ndelay2]
                    outbuf[i1] = itemp
                    i2 += 1
    return

"""This function bit-reverses the given value "inval" with the number of
bits, "nbits".    ----  R. Ramachandran, 10-Nov-97, nfra.
python version ----  H. Chen   Modified 2014
cython version ---- E. Enriquez + P.Schellart 1-Feb-2016
"""
cpdef int bitrev(int inval, int nbits) except *:
    cdef int ibitr
    cdef int k
    cdef int ifact

    if nbits <= 1:
        ibitr = inval
    else:
        ifact = 1
        for i in range(1, nbits):
           ifact *= 2
        k = inval
        ibitr = (1 & k) * ifact
        for i in range(2, nbits+1):
            k /= 2
            ifact /= 2
            ibitr += (1 & k) * ifact
    return ibitr




#!/usr/bin/env python

import numpy as np
import logging
logger_hf = logging.getLogger(__name__)


def chan_freq(header, fine_channel, tdwidth, ref_frame):
    """

    Args:
      header: 
      fine_channel: 
      tdwidth: 
      ref_frame: 

    Returns:

    """
    fftlen = header['NAXIS1']
    chan_index = fine_channel - (tdwidth-fftlen)/2
    chanfreq = header['FCNTR'] + (chan_index - fftlen/2)*header['DELTAF']
    #/* apply doppler correction */
    if ref_frame == 1:
        chanfreq = (1 - header['baryv']) * chanfreq
    return chanfreq


def bitrev(inval, nbits):
    """This function bit-reverses the given value "inval" with the number of
    bits, "nbits".    ----  R. Ramachandran, 10-Nov-97, nfra.
    python version ----  H. Chen   Modified 2014

    Args:
      inval: number to be bit-reversed
      nbits: The length of inval in bits. If user only wants the bit-reverse of a certain amount of bits of
    inval, nbits is the amount of bits to be reversed counting from the least significant (rightmost)
    bit. Any bits beyond this length will not be reversed and will be truncated from the result.

    Returns:
      : the bit-reverse of inval. If there are more significant bits beyond nbits, they are truncated.

    """
    if nbits <= 1:
        ibitr = inval
    else:
        ifact = 1
        for i in range(1, nbits):
           ifact *= 2
        k = inval
        ibitr = (1 & k) * ifact
        for i in range(2, nbits+1):
            k = int(k / 2)
            ifact = int(ifact / 2)
            ibitr += (1 & k) * ifact
    return ibitr


def FlipX(outbuf, xdim, ydim):
    """This function takes in an array of values and iteratively flips ydim chunks of values of length xdim. For example,
    if you have an array [1, 2, 3, 4, 5, 6, 7, 8, 9, 10] and enter it with xdim = 5 and ydim = 2, the array will be
    modified to become [5, 4, 3, 2, 1, 10, 9, 8, 7, 6]. Note that if you wish for the whole array to be modified in this
    way, xdim * ydim should equal the length of the array. If ydim * xdim is greater than the length of the array, this
    function will error.

    Args:
      outbuf: ndarray,        an array with shape = (int, 1)
      xdim: int,            size of segments to be flipped
      ydim: int,            amount of segments of size xdim to be flipped

    Returns:

    """
    temp = np.empty_like(outbuf[0:xdim])
    logger_hf.debug("FlipX: temp array dimension: %s"%str(temp.shape))

    for j in range(0, ydim):
        indx = j * xdim
        np.copyto(temp, outbuf[indx:indx+xdim])
        np.copyto(outbuf[indx: indx+xdim], temp[::-1])
    return


def comp_stats(arrey):
    """Compute mean and stddev of floating point vector array in a fast way, without using the outliers.

    Args:
      arrey: ndarray,        floating point vector array

    Returns:
      : float, float,   median and standard deviation of input array

    """

    new_vec = np.sort(arrey,axis=None)

    #Removing the lowest 5% and highest 5% of data, this takes care of outliers.
    new_vec = new_vec[int(len(new_vec)*.05):int(len(new_vec)*.95)]
    the_median = np.median(new_vec)
    the_stddev = new_vec.std()

    return the_median, the_stddev



#!/usr/bin/env python
"""

This module is very BL-GBT oriented. Will transfer to different repo.

 ..author: Emilio Enriquez
"""

import pandas as pd
from argparse import ArgumentParser
import matplotlib.pylab as plt
import socket
import numpy as np

pd.options.mode.chained_assignment = None  # To remove pandas warnings: default='warn'

def find_scan_sets(filename,band,ok_bands = ['L','S']):
    """Finding scan sets, needs filename of csv file and band.

    Args:
      filename: 
      band: 
      ok_bands:  (Default value = ['L')
      'S']: 

    Returns:

    """

    if '.csv' in filename:
        spider_danny = True
    else:
        spider_danny = False

    #---------------------------
    # Read in the full "A list" of stars
    # This comes from the BL database.
    #---------------------------
    local_host = socket.gethostname()

    #To Do:
    if 'bl' in local_host:
        master_file = open('/home/obs/logs/target_list_5-50pc.lst')
    else:
        master_file = open('/Users/jeenriquez/RESEARCH/SETI_BL/L_band/target_list_5-50pc.lst')

    a_list_master = master_file.read().splitlines()

    #---------------------------
    # Find the good  data by:
    # - Reading in the output from spider.
    #---------------------------
    # Initial data frame set up

    if spider_danny:
        try:
            df2 = pd.read_csv(filename) # New style spider files created by Danny's code.
            #,Unnamed: 0,az_start,data_type,fch1,filepath,filesize,foff,host,ibeam,machine_id,nbeams,nbits,nchans,nifs,nints,rawdatafile,source_name,src_dej,src_raj,telescope_id,tsamp,tstart
        except:
            IOError('Error opening file: %s'%filename)

        #column_names
        file = 'filepath'
        freq_chan1 = 'fch1'
        nchans = 'nchans'
        source_name = 'source_name'
        nints = 'nints'
        tstart = 'tstart'

        #Selection of high resolution data
        df3 = df2[df2[file].str.contains("gpuspec.0000.h5",na=False)]

    else:
        try:
            df = pd.read_csv(filename, sep=",|=", header=None,engine='python')    # Old style spider files.
        except:
            IOError('Error opening file: %s'%filename)

        #Reorganizing
        df2 = df.ix[:,1::2] #data
        df2.columns = list(df.ix[0,0::2]) #header

        #column_names
        file = 'file'
        freq_chan1 = 'Frequency of channel 1 (MHz)'
        nchans = 'Number of channels'
        source_name = 'Source Name'
        nints = 'Number of samples'
        tstart = 'Time stamp of first sample (MJD)'

        #Selection of high resolution data
        df3 = df2[df2[file].str.contains("gpuspec.0000.fil",na=False)]

    #Selection of observations from given band (soft)
    if band == 'S':
        df3 = df3[((df3[freq_chan1] > 2400.) & (df3[freq_chan1] < 3500.))]
    elif band == 'L':
        df3 = df3[df3[freq_chan1] < 2500]
    else:
        raise ValueError('Please probide one of the available bands:' + ok_bands)

    #---------------------------
    # Adding some extra columns for later look at the good set of data.

    df3['bands_used'] = [df3[file][ii].split('/')[-1].split('_')[1].replace('blc','') for ii in df3.index]
    df3['mid_Freq'] = df3[freq_chan1]-2.7939677238464355e-06*df3[nchans]/2.
    if band == 'L':
        df3['mid_Freq2'] = df3[freq_chan1]-2.835503418452676e-06*df3[nchans]/2.

    df3[source_name] = df3[source_name].str.upper()

    #---------------------------
    #Check the data that has the good bands.
    if band == 'S':
        df3 = df3[df3['bands_used'].str.contains('.1.2.3.4.5.6')]  # The dots are use as while cards (*)
    elif band == 'L':
        df3 = df3[df3['bands_used'].str.contains('.2.3.4.5')]
    else:
        raise ValueError('Please probide one of the available bands:' + ok_bands)

    #---------------------------
    #Check for correct Number of Samples
    df3 = df3[df3[nints] == 16]

    #---------------------------
    #Check for high resolution data with the bad central frequency.
    # The two main frequency resolutions used are -2.7939677238464355e-06 and -2.835503418452676e-06  .

    if band == 'L':
        # Ok mid Freqs around 1501.4648, or 1475.0976

        df_good_mid_FreqA = df3[((df3['mid_Freq'] > 1501.4) & (df3['mid_Freq'] < 1501.5))]
        df_good_mid_FreqB = df3[((df3['mid_Freq2'] > 1501.4) & (df3['mid_Freq2'] < 1501.5))]

        df_good_mid_FreqA2 = df3[((df3['mid_Freq'] > 1475.0) & (df3['mid_Freq'] < 1475.1))]
        df_good_mid_FreqB2 = df3[((df3['mid_Freq2'] > 1475.0) & (df3['mid_Freq2'] < 1475.1))]

        #bad mid freqs?

        #concat
        df3 = pd.concat([df_good_mid_FreqA,df_good_mid_FreqA2,df_good_mid_FreqB,df_good_mid_FreqB2])

    elif band == 'S':
        # Ok mid Freqs around 2300.3906, 2269.6289, 2307.75
        df3 = df3[((df3['mid_Freq'] > 2269.) & (df3['mid_Freq'] < 2310.))]
        #bad mid freqs 2276.23
        df3 = df3[~((df3['mid_Freq'] > 2276.2) & (df3['mid_Freq'] < 2276.3))]
    else:
        raise ValueError('Please probide one of the available bands:' + ok_bands)

    #---------------------------
    # Apply format change in tstart and sort by it.

    df3[tstart] = df3[tstart].apply(pd.to_numeric)
    df3 = df3.sort_values(by=[tstart])

    #---------------------------
    # Selecting only the targets in the A-list

    #Selecting all the targets from the B list
    df_targets_blist = df3[~df3[source_name].isin(a_list_master)]
    df_targets_clist =  df3[df3[source_name].str.contains('_OFF',na=False)]
    df_targets_blist = pd.concat([df_targets_blist,df_targets_clist])
    else_list = df_targets_blist[source_name].unique()

    #Selecting all the good targets from the A list
    df_targets_alist = df3[~df3[source_name].isin(else_list)]

    #---------------------------
    #Showing some info

    print('------      o      --------')
    a_unique = df_targets_alist[source_name].unique()
    print('This list was created for the %s band data'%band)
    print('The total number of targets from the A-list that:')
    print('Observed and spliced is      : %i'%(len(a_unique)))

    #---------------------------
    # Group the df_targets and look for the ones observed 3 times or more
    # Grouping without date constrains.
    df_targets = df_targets_alist.groupby(source_name).count()[file] > 2
    df_bool_list = df_targets.tolist()
    list_completed = list(df_targets[df_bool_list].index.values)

    #---------------------------
    #Selecting targets with "completed" observations
    df_targets_alist = df_targets_alist[df_targets_alist[source_name].isin(list_completed)]
    alist_completed_unique = df_targets_alist[source_name].unique()

    print('Have at least 3 observations : %i'%(len(alist_completed_unique)))

    #---------------------------
    #Creating list of targets

    list_targets =''
    list_A_stars=[]
    list_A_times=[]

    counting = 0



    for a_star in alist_completed_unique:

        df_a_star = df_targets_alist[df_targets_alist[source_name] == a_star]

#         if len(df_a_star[tstart]) > 5:
#             print 'WARNING: Too many observations of target: ', a_star
#
#         mid_time = df_a_star[tstart].median()
#         df_a_star['delta_t'] = df_a_star[tstart].apply(lambda x: float(x) - float(mid_time))
#
#         #Taking observations only from the same day.
#         df_a_star = df_a_star[df_a_star['delta_t'] < 1.]

        list_a_star_times = df_a_star[tstart].unique()
#        scan_count = 0

        for a_time in list_a_star_times:

            #Calculating delta t
            df_a_star['delta_t'] = df_a_star[tstart].apply(lambda x: np.abs(float(x) - float(a_time)))
            time_list = df_a_star[df_a_star['delta_t'] < 0.05][tstart]

            if  len(time_list)>3:

                if np.isin(time_list,list_A_times).sum() > 2:
                    print('WARNING: Skiping this A observation:', a_star,a_time, 'Length is :', len(time_list))
                    continue

            #Taking observations only with 3 nearby.
#             This fails when having both many observations of one target, but some
#             if len(time_list)>3:
#                 scan_count+=1
#                 if scan_count > 3:
#                     scan_count = 0
#                     print 'WARNING: Skiping this A observation:', a_star,a_time, 'Length is :', len(time_list)
#                     continue
#
# #                 mid_time = time_list.median()
# #                 time_diffs = np.abs((time_list- mid_time).unique())
# #
# #                 if np.abs(a_time - mid_time) > time_diffs[2]+.0001:
# #                     print 'WARNING: Skiping this A observation:', a_star,a_time, 'Length is :', len(time_list)
# #                     continue


            elif  len(time_list)<3:
                continue

            #Making tmp DF for ONs and OFFs
            df_tmp = df3[ (df3[tstart] > float(a_time)-0.1) & (df3[tstart] < float(a_time)+0.1)]
            df_tmp['delta_t'] = df_tmp[tstart].apply(lambda x: float(x) - float(a_time))

            try:
                ii = df_tmp[df_tmp['delta_t']>0.001]['delta_t'].idxmin()   #Find B star index  #.001 = 1.44 min
                #jj = df_tmp[df_tmp['delta_t']>=0]['delta_t'].idxmin()   #Find A star index

                #full path
                a_name = list(df_a_star[df_a_star[tstart] == a_time][file])[0]
                b_name = df_tmp[file][ii]

            except:
                print('WARNING: Could not find ON-OFF pair: ', a_star, a_name)

                #full path
                a_name = list(df_a_star[df_a_star[tstart] == a_time][file])[0]
                b_name = 'Place_holder_to_delete.'   # Creates a place holder in the list of obs for manual deletion.

#                continue

            if a_name == b_name:
                print('WARNING: Skipping (a=b). ', a_name)
#                continue

            #Find if data pairs are not in the same node (thus 'non-co-living').
#            if a_name.split('/')[1] != b_name.split('/')[1]:
#                print 'WARNING: A and B not in same location.', a_name
#                 continue
#             else:
            #a_star_file_name, b_star_file_name
            tmp_string = ['/mnt_'+local_host+a_name,'\n','/mnt_'+local_host+b_name]
            list_targets += ''.join(tmp_string)+'\n'
            counting+=1

            list_A_times.append(a_time)



        list_A_stars.append(a_star+'\n')

    print('Actual number of stars at least one complete set: %i'%(len(list_A_stars)))

    #---------------------------
    #Save lists

    with open('%s_band_target_pairs.lst'%band,'w') as file_list:
    #with open('/datax/users/eenriquez/L-band_analysis/'+'%s_band_target_pairs.lst'%band,'w') as file_list:
        file_list.write(list_targets)

    with open('%s_band_A_stars.lst'%band,'w') as file_list:
    #with open('/datax/users/eenriquez/L-band_analysis/'+'%s_band_A_stars.lst'%band,'w') as file_list:
        file_list.write(''.join(list_A_stars))

def main():
    """Main funtion for find_event scripts."""

    parser = ArgumentParser(description="Command line utility for creating a list of ON/OFF pairs of spliced data.")
    parser.add_argument('filename', type=str, help='Full path and filename to read (csv).')
    parser.add_argument('band', type=str,default='L', help='Which band to use? (L,S)')
    args = parser.parse_args()

    #Available bands
    ok_bands = ['L','S']

    if args.band not in ok_bands:
        raise ValueError('Please probide one of the available bands:' + ok_bands)
    else:
        band = args.band

    filename = args.filename

    find_scan_sets(filename,band,ok_band=ok_bands)

if __name__ == "__main__":
    main()




#!/usr/bin/env python

import os
import numpy as np
import math
from pkg_resources import resource_filename
from pkg_resources import resource_filename
from blimpy import Waterfall
from blimpy.io import sigproc
import h5py

import logging
logger = logging.getLogger(__name__)

#For debugging
#import cProfile
#import pdb;# pdb.set_trace()

SIZE_LIM = 256.0   # File size limit in MB. If larger then make a split mapping.

class DATAHandle:
    """Class to setup input file for further processing of data. Handles conversion to h5 (from fil), extraction of
    coarse channel info, waterfall info, and file size checking.

    Args:

    Returns:

    """
    def __init__(self, filename=None, size_limit=SIZE_LIM, out_dir='./', n_coarse_chan=None, coarse_chans=None):
        """
        :param filename:        string,      name of file (.h5 or .fil)
        :param size_limit:      float,       maximum size in MB that the file is allowed to be
        :param out_dir:         string,      directory where output files should be saved
        """

        if filename and os.path.isfile(filename):
            self.filename = filename
            self.out_dir = out_dir
            self.n_coarse_chan = n_coarse_chan
            self.coarse_chans = coarse_chans

            if not h5py.is_hdf5(filename):
                if not sigproc.is_filterbank(filename):
                    raise IOError('No correct format, need .h5. Try again...')
                else:
                    logger.info("File .fil detected. Attempting to create .h5 file in current directory...")
                    try:
                        self.__make_h5_file()
                    except:
                        raise IOError('Unable to create .h5 file. Please, try again with correct format.')

            self.filestat = os.stat(filename)
            self.filesize = self.filestat.st_size/(1024.0**2)

            # Make sure file is not larger than limit. If it is, we must split the file
            if self.filesize > size_limit:
                logger.info("The file is of size %f MB, exceeding our size limit %f MB. Split needed..."%(self.filesize, size_limit))
                self.data_list = self.__split_h5()
            else:
                #EE This is here mainly for testing. Since need to keep in mind the band pass shape.
                logger.debug("File size %f MB within range %f MB, okay..."%(self.filesize, size_limit))
                data_obj = DATAH5(filename)
                self.data_list = [data_obj]

            self.status = True

        else:
            self.status = False
            errmsg = "File {} doesn\'t exist, please check!".format(filename)
            logger.error(errmsg)
            raise IOError(errmsg)

    def get_info(self):
        """:return:    dict,    header of the blimpy file"""

        fil_file = Waterfall(self.filename,load_data=False)
        return fil_file.header

    def __make_h5_file(self,):
        """
        Converts file to h5 format. Saves output in current dir. Sets the filename attribute of the calling DATAHandle
        to the (new) filename.
        :return: void
        """

        fil_file = Waterfall(self.filename)
        bn = os.path.basename(self.filename)
        new_filename = os.path.join(self.out_dir, bn.replace('.fil', '.h5'))
        fil_file.write_to_hdf5(new_filename)
        self.filename = new_filename

    def __split_h5(self, size_limit=SIZE_LIM):
        """
        Creates a plan to select data from single coarse channels.
        :param size_limit:      float,              maximum size in MB that the file is allowed to be
        :return:                list[DATAH5],       the list contains a DATAH5 object for each of the coarse channels in
                                                    the file
        """

        data_list = []

        #Instancing file.
        try:
            fil_file = Waterfall(self.filename)
        except:
            logger.error("Error encountered when trying to open file: %s"%self.filename)
            raise IOError("Error encountered when trying to open file: %s"%self.filename)

        #Finding lowest freq in file.
        f_delt = fil_file.header['foff']
        f0 = fil_file.header['fch1']

        #Looping over the number of coarse channels.
        if self.n_coarse_chan is not None:
            n_coarse_chan = self.n_coarse_chan
        elif fil_file.header.get('n_coarse_chan', None) is not None:
            n_coarse_chan = fil_file.header['n_coarse_chan']
        else:
            n_coarse_chan = int(fil_file.calc_n_coarse_chan())

        # Only load coarse chans of interest -- or do all if not specified
        if self.coarse_chans in (None, ''):
            self.coarse_chans = range(n_coarse_chan)

        for chan in self.coarse_chans:

            #Calculate freq range for given course channel.
            f_start = f0 + chan*(f_delt)*fil_file.n_channels_in_file/n_coarse_chan
            f_stop = f0 + (chan+1)*(f_delt)*fil_file.n_channels_in_file/n_coarse_chan

            if f_start > f_stop:
                f_start, f_stop = f_stop, f_start

            data_obj = DATAH5(self.filename, f_start=f_start, f_stop=f_stop, coarse_chan=chan, tn_coarse_chan=n_coarse_chan)

#----------------------------------------------------------------

            #This appends to a list of all data instance selections. So that all get processed later.
            data_list.append(data_obj)

        return data_list

class DATAH5:
    """This class is where the waterfall data is loaded, as well as the header info.
    It creates other attributes related to the dedoppler search (load_drift_indexes).

    Args:

    Returns:

    """

    def __init__(self, filename, size_limit = SIZE_LIM,f_start=None, f_stop=None,t_start=None, t_stop=None,coarse_chan=1,tn_coarse_chan=None):
        """
        :param filename:        string      name of file
        :param size_limit:      float       maximum size in MB that the file is allowed to be
        :param f_start:         float       start frequency in MHz
        :param f_stop:          float       stop frequency in MHz
        :param t_start:         int         start integration ID
        :param t_stop:          int         stop integration ID
        :param coarse_chan:     int
        :param tn_coarse_chan:  int
        """

        self.filename = filename
        self.closed = False
        self.f_start = f_start
        self.f_stop = f_stop
        self.t_start = t_start
        self.t_stop = t_stop
        self.tn_coarse_chan = tn_coarse_chan

        #Instancing file.
        try:
            self.fil_file = Waterfall(filename,f_start=self.f_start, f_stop=self.f_stop,t_start=self.t_start, t_stop=self.t_stop,load_data=False)
        except:
            logger.error("Error encountered when trying to open file %s"%filename)
            raise IOError("Error encountered when trying to open file %s"%filename)

        #Getting header
        try:
            if self.tn_coarse_chan:
                header = self.__make_data_header(self.fil_file.header,coarse=True)
            else:
                header = self.__make_data_header(self.fil_file.header)
        except:
            logger.debug('The fil_file.header is ' % self.fil_file.header)
            raise IOError("Error accessing header from file: %s." % self.filename)

        self.header = header

        self.fftlen = header['NAXIS1']
 
        #EE To check if swapping tsteps_valid and tsteps is more appropriate.
        self.tsteps_valid = header['NAXIS2']
        self.tsteps = int(math.pow(2, math.ceil(np.log2(math.floor(self.tsteps_valid)))))
 
        self.obs_length = self.tsteps_valid * header['DELTAT']
        self.drift_rate_resolution = (1e6 * np.abs(header['DELTAF'])) / self.obs_length   # in Hz/sec
        self.header['baryv'] = 0.0
        self.header['barya'] = 0.0
        self.header['coarse_chan'] = coarse_chan
        
        #EE For now I'm not using a shoulder. This is ok as long as I'm analyzing each coarse channel individually.
        #EE In general this parameter is an integer (even number).
        #This gives two regions, each of n*steps, around spectra[i]
        self.shoulder_size = 0
        self.tdwidth = self.fftlen + self.shoulder_size*self.tsteps

    def load_data(self,):
        """Read the spectra and drift indices from file.
        :return:    ndarray, ndarray        spectra, drift indices

        Args:

        Returns:

        """
        self.fil_file.read_data(f_start=self.f_start, f_stop=self.f_stop)

        #Blanking DC bin.
        n_coarse_chan = int(self.fil_file.calc_n_coarse_chan())
        if n_coarse_chan != self.fil_file.calc_n_coarse_chan():
            logger.warning('The file/selection is not an integer number of coarse channels. This could have unexpected consequences. Let op!')
        self.fil_file.blank_dc(n_coarse_chan)

        spec = np.squeeze(self.fil_file.data)
        spectra = np.array(spec, dtype=np.float64)

        # DCP APR 2020 -- COMMENTED OUT. THIS IS BREAKING STUFF IN CURRENT VERSION.
        #Arrange data in ascending order in freq if not already in that format.
        #if self.header['DELTAF'] < 0.0:
        #    spectra = spectra[:,::-1]

        #This check will add rows of zeros if the obs is too short (and thus not a power of two rows).
        if spectra.shape[0] != self.tsteps:
            spectra = np.append(spectra, np.zeros((self.tsteps-spectra.shape[0], self.fftlen)), axis=0)
        self.tsteps_valid = self.tsteps
        self.obs_length = self.tsteps * self.header['DELTAT']

        if spectra.shape != (self.tsteps_valid, self.fftlen):
            logger.error('Something is wrong with array size.')
            raise IOError('Something is wrong with array size.')

        drift_indexes = self.load_drift_indexes()

        return spectra, drift_indexes

    def load_drift_indexes(self):
        """The drift indices are read from a stored file so that there is no need to recalculate. This speed things up.
        :return:    ndarray     drift indices

        Args:

        Returns:

        """
        n = int(np.log2(self.tsteps))
        di_array = np.genfromtxt(resource_filename('turbo_seti', 'drift_indexes/drift_indexes_array_%d.txt'%n), delimiter=' ', dtype=int)

        ts2 = int(self.tsteps/2)
        drift_indexes = di_array[(self.tsteps_valid - 1 - ts2), 0:self.tsteps_valid]
        return drift_indexes

    def __make_data_header(self,header,coarse=False):
        """
        Takes header into fits header format
        :param header:      dict        blimpy waterfall header
        :param coarse:      Boolean     whether or not there are coarse channels to analyze
        :return:            dict
        """

        base_header = {}

        #used by file_writers.py
        base_header['SOURCE'] = header['source_name']
        base_header['MJD'] = header['tstart']
        base_header['DEC'] = str(header['src_dej'])
        base_header['RA'] = str(header['src_raj'])
        base_header['DELTAF'] =  header['foff']
        base_header['DELTAT'] = float(header['tsamp'])

        #used by helper_functions.py
        if coarse:
            base_header['NAXIS1'] = int(header['nchans']/self.tn_coarse_chan)
            base_header['FCNTR'] = np.abs(self.f_stop - self.f_start)/2. + np.min(self.f_start, self.f_stop)
        else:
            base_header['NAXIS1'] = int(header['nchans'])
            base_header['FCNTR'] = float(header['fch1']) + header['foff'] * base_header['NAXIS1'] / 2

        #other header values.
        base_header['NAXIS'] = 2
        base_header['NAXIS2'] = int(self.fil_file.n_ints_in_file)
        return base_header

    def close(self):
        """Closes file and sets the data attribute `.closed` to
        True. A closed object can no longer be used for I/O operations.
        `close()` may be called multiple times without error.

        Args:

        Returns:

        """

        # Call file object destructor which should close the file
        del self.fil_file

        self.closed = True




#!/usr/bin/env python

"""
Front-facing script to find drifting, narrowband events in a set of generalized 
cadences of ON-OFF radio SETI observations.

In this code, the following terminology is used:
Hit = single strong narrowband signal in an observation
Event = a strong narrowband signal that is associated with multiple hits
        across ON observations
    
The main function contained in this file is *find_event_pipeline*
    Find_event_pipeline calls find_events from find_events.py to read a list 
    of turboSETI .dat files. It then finds events within this group of files. 
                      
Usage (beta):
    import find_event_pipeline;
    find_event_pipeline.find_event_pipeline(dat_file_list_str, 
                                            SNR_cut=10,
                                            check_zero_drift=False,
                                            filter_level=3, 
                                            on_off_first='ON', 
                                            number_in_cadence=6, 
                                            saving=True,  
                                            user_validation=False)
    
    dat_file_list_str   The string name of a plaintext file ending in .lst 
                        that contains the filenames of .dat files, each on a 
                        new line, that were created with seti_event.py. The 
                        .lst should contain a set of cadences (ON observations 
                        alternating with OFF observations). The cadence can be 
                        of any length, given that the ON source is every other 
                        file. This includes Breakthrough Listen standard ABACAD
                        as well as OFF first cadences like BACADA. Minimum 
                        cadence length is 2, maximum cadence length is 
                        unspecified (currently tested up to 6).
                                   
                        Example: ABACAD|ABACAD|ABACAD
                   
    SNR_cut             The threshold SNR below which hits in the ON source 
                        will be disregarded. For the least strict thresholding, 
                        set this parameter equal to the minimum-searched SNR 
                        that you used to create the .dat files from 
                        seti_event.py. Recommendation (and default) is 10.
                   
    check_zero_drift    A True/False flag that tells the program whether to
                        include hits that have a drift rate of 0 Hz/s. Earth-
                        based RFI tends to have no drift rate, while signals
                        from the sky are expected to have non-zero drift rates.
                        Default is False.
                        
    filter_threshold    Specification for how strict the hit filtering will be.
                        There are 3 different levels of filtering, specified by
                        the integers 1, 2, and 3. Filter_threshold = 1 
                        returns hits above an SNR cut, taking into account the
                        check_zero_drift parameter, but without an ON-OFF check.
                        Filter_threshold = 2 returns hits that passed level 1
                        AND that are in at least one ON but no OFFs. 
                        Filter_threshold = 3 returns events that passed level 2
                        AND that are present in *ALL* ONs. Default is 3.
                        
    on_off_first        Tells the code whether the .dat sequence starts with
                        the ON or the OFF observation. Valid entries are 'ON'
                        and 'OFF' only. Default is 'ON'.
    
    number_in_cadence   The number of files in a single ON-OFF cadence.
                        Default is 6 for ABACAD.
                        
    saving              A True/False flag that tells the program whether to 
                        save the output array as a .csv
                        
    user_validation     A True/False flag that, when set to True, asks if the
                        user wishes to continue with their input parameters
                        (and requires a 'y' or 'n' typed as confirmation)
                        before beginning to run the program. Recommended when
                        first learning the program, not recommended for 
                        automated scripts.
                    
author: 
    Version 2.0 - Sofia Sheikh (ssheikhmsa@gmail.com), 
    Version 1.0 - Emilio Enriquez (jeenriquez@gmail.com)

***
NOTE: This code works for .dat files that were produced by seti_event.py
after turboSETI version 0.8.2, and blimpy version 1.1.7 (~mid 2019). The 
drift rates *before* that version were recorded with the incorrect sign
and thus the drift rate sign would need to be flipped in the make_table 
function.
***

"""

#required packages and programs
import find_event
import pandas as pd

#required for updated_find_event
import time
import numpy as np

def find_event_pipeline(dat_file_list_str,
                        SNR_cut=10, 
                        check_zero_drift=False, 
                        filter_threshold=3, 
                        on_off_first='ON', 
                        number_in_cadence=6, 
                        saving=True, 
                        user_validation=False): 
    """

    Args:
      dat_file_list_str: 
      SNR_cut:  (Default value = 10)
      check_zero_drift:  (Default value = False)
      filter_threshold:  (Default value = 3)
      on_off_first:  (Default value = 'ON')
      number_in_cadence:  (Default value = 6)
      saving:  (Default value = True)
      user_validation:  (Default value = False)

    Returns:

    """
    print()
    print("************   BEGINNING FIND_EVENT PIPELINE   **************")
    print()
    print("Assuming the first observation is an " + on_off_first)
    
    #Opening list of files
    dat_file_list = open(dat_file_list_str).readlines()
    dat_file_list = [files.replace('\n','') for files in dat_file_list]
    dat_file_list = [files.replace(',','') for files in dat_file_list]
    n_files = len(dat_file_list)
    
    print("There are " + str(len(dat_file_list)) + " total files in the filelist " + dat_file_list_str)
    print("therefore, looking for events in " + str(int(n_files/number_in_cadence)) + " on-off set(s)")
    print("with a minimum SNR of " + str(SNR_cut))
    
    if filter_threshold == 1:
        print("Present in an A source only, above SNR_cut")
    if filter_threshold == 2:
        print("Present in at least one A source with RFI rejection from the off-sources")
    if filter_threshold == 3:
        print("Present in all A sources with RFI rejection from the off-sources")
    
    if check_zero_drift == False:
        print("not including signals with zero drift")
    if check_zero_drift == True:
        print("including signals with zero drift")
    if saving == False:
        print("not saving the output files")
    if saving == True:
        print("saving the output files")    
    
    if user_validation == True:
        question = "Do you wish to proceed with these settings?"
        while "the answer is invalid":
            reply = str(input(question+' (y/n): ')).lower().strip()
            if reply == '':
                return
            if reply[0] == 'y':
                break
            if reply[0] == 'n':
                return
    
    #Looping over number_in_cadence chunks.
    candidate_list = []
    for i in range((int(n_files/number_in_cadence))):
        file_sublist = dat_file_list[number_in_cadence*i:((i*number_in_cadence)+(number_in_cadence))]
        if on_off_first == 'ON':
            name=file_sublist[0].split('_')[5]  
        if on_off_first == 'OFF':
            name=file_sublist[1].split('_')[5] 
        print()
        print("***       " + name + "       ***")
        print()
        cand = find_event.find_events(file_sublist, 
                                      SNR_cut=SNR_cut, 
                                      check_zero_drift=check_zero_drift, 
                                      filter_threshold=filter_threshold, 
                                      on_off_first=on_off_first)
        cand_len = 1
        if cand is None:
            cand_len = 0
        if cand_len != 0:
            candidate_list.append(cand)
    if len(candidate_list) > 0:
        find_event_output_dataframe = pd.concat(candidate_list)
    else:
        "Sorry, no potential candidates with your given parameters :("
        find_event_output_dataframe = []

    print("************  ENDING FIND_EVENT PIPELINE   **************")
    
    if saving == True:
        if check_zero_drift == True:
            filestring = name + '_f' + str(filter_threshold) + '_snr' + str(SNR_cut) + '_zero' + '.csv'
        else:
            filestring = name + '_f' + str(filter_threshold) + '_snr' + str(SNR_cut) + '.csv'
        
        find_event_output_dataframe.to_csv(filestring)


    return(find_event_output_dataframe)

#***********plotting code*********
#name_array = np.array(['HD_109376', 'HD_115153', 'HD_152956','LTT_7183', 'LTT_7494', 'GJ_757', 'BHIP_61711', 'HD_117004', 'LTT_6827','BHIP_88982', 'BHIP_93072', 'BHIP_96440'])
#    find_event_output_dataframe.to_csv(r'/home/ssheikh/ETZ_analysis/find_event_pipes/' + node_string + '_f' + str(filter_level) + '.csv')
#    print(".csv successfully saved")
#    trimmed_event_dataframe = find_event_output_dataframe.loc[:, ['Source', 'Freq', 'DriftRate','FreqStart']]
#    if len(find_event_output_dataframe) > 100:
#        print("too many events to plot, setting a stricter SNR threshold")
#        if node_string == 'blc24':
#            trimmed_event_dataframe = find_event_output_dataframe.loc[find_event_output_dataframe['SNR'] > 50, ['Source', 'Freq', 'DriftRate','FreqStart']]
#         
#        if node_string == 'blc25':
#            trimmed_event_dataframe = find_event_output_dataframe.loc[find_event_output_dataframe['SNR'] > 100, ['Source', 'Freq', 'DriftRate','FreqStart']]
#           
#        if node_string == 'blc34':
#                        trimmed_event_dataframe = find_event_output_dataframe.loc[find_event_output_dataframe['SNR'] > 20, ['Source', 'Freq', 'DriftRate','FreqStart']]
#           
#        print("now less than 100 events - plotting...")
#
#    #knowing whether a star is A or B, and what its indexes are
#    ab_array = np.zeros(len(name_array))
#    ab_array[:(len(name_array)/2)] = 1
#    index_array = range(0, len(dat_file_list), 6)
#
#    #Opening list of files, make them A and B lists
#    fil_file_list = open(fil_file_list_string).readlines()
#    fil_file_list = [files.replace('\n','') for files in fil_file_list]
#    fil_file_list = [files.replace(',','') for files in fil_file_list]
#    fil_file_list_b = fil_file_list[1:]
#    fil_file_list_b.append(fil_file_list[0])
#    fil_file_list = fil_file_list + fil_file_list_b

#    #loop through all target stars with promising candidates and plot them
#    for source in trimmed_event_dataframe.Source.unique():
#        print(source)
#        star_id_index = np.where(name_array == source)[0][0]
#        start_index = index_array[star_id_index]
        #plotting
#        plot_candidate_events_individually(trimmed_event_dataframe, correct_fils, source, node_string, filter_level, show=False, overwrite=False, offset=0)


#event_finder_and_plotter('blc00', 3, 10, #'/home/ssheikh/ETZ_analysis/sixes/A_lists/dats/blc/blc00_dats.lst', #'/home/ssheikh/ETZ_analysis/sixes/A_lists/fils/blc/blc00_fils.lst')





#!/usr/bin/env python

import numpy as np
import astropy.io.fits as pyfits
from .helper_functions import chan_freq


import logging

def tophits_writer(spectra_slice, hit_indices, header, format='txt'):
    """

    Args:
      spectra_slice: 
      hit_indices: 
      header: 
      format: (Default value = 'txt')

    Returns:

    """
    return None

class GeneralWriter:
    """Wrapper class for file operations."""
    def __init__(self, filename='', mode='a'):
        """
        Initializes GeneralWriter object. Opens given file with given mode, sets new object's filehandle to the file
        object, sets the new object's filename to the file's name, then closes the file.

        Args:
            filename:    string,     name of file on which we would like to perform operations
            mode:        string,     mode which we want to use to open file, same modes as the built-in python
                                      built-in open function: r - read, a - append, w -write, x - create
        """
        with open(filename, mode) as myfile:
            self.filehandle = myfile
            self.filename = filename
        return None

    def close(self):
        """Closes file object if it is open.
        """
        if self.filehandle.closed:
            pass
        else:
            self.filehandle.close()

    def open(self, mode='a'):
        """Opens the file with the inputted mode, then closes it. Does not actually leave the file opened, only used for
        changing mode.

        Args:
          mode: string,     mode which we want to assign to this file, same modes as the built-in python
        built-in open function: r - read, a - append, w -write, x - create (Default value = 'a')
        """
        if self.filehandle.closed:
            with open(self.filename, mode) as myfile:
                self.filehandle = myfile
        elif self.filehandle.mode == mode:
            return
        else:
            self.close()
            with open(self.filename, mode) as myfile:
                self.filehandle = myfile

    def is_open(self):
        """Checks if file is open.
        Returns:  boolean,    true if file is open, false otherwise
        """
        return not self.filehandle.closed

    def writable(self):
        """Checks if file is open, and if it is, checks that mode is either write or append.
        :return:    boolean,    true if file is open and writeable, false otherwise

        Args:

        Returns:

        """
        if self.is_open() and (('w' in self.filehandle.mode) or ('a' in self.filehandle.mode)):
            return True
        else:
            return False

    def write(self, info_str, mode='a'):
        """Sets file mode to a writeable mode and opens it if it is not already open in a writeable mode, writes info_str
        to it, and then closes it. If the file was not previously open when this is called, the file is closed after
        writing in order to maintain the state the filewriter was in before.

        Args:
          info_str: string,     data to be written to file
          mode: string,     mode for file. If it is not a writeable mode, it will be set to a writeable mode (Default value = 'a')

        Returns:

        """
        if (not 'w' in mode) and (not 'a' in mode):
            mode = 'a'
        if not self.writable():
            with open(self.filename, mode) as myfile:
                myfile.write(info_str)
                self.filehandle = myfile
        else:
            self.filehandle.write(info_str)

    def start_over(self):
        """Wipes the given file and then opens it in append mode. Note that opening a file in write mode wipes it, so we
        don't have to actually write an empty string to it.

        Args:

        Returns:

        """
        self.open('w')
        self.write('')
        self.open('a')

class FileWriter(GeneralWriter):
    """Used to write information to turboSETI output files."""
    def __init__(self, filename, header):
        """ Initializes FileWriter object and writes its header.

        :param filename:    string,     name of file on which we would like to perform operations
        :param header:      dict,       information to be written to header of file filename
        """
        GeneralWriter.__init__(self, filename)
        self.write('# -------------------------- o --------------------------\n')
        self.write('# File ID: %s \n'%(filename.split('/')[-1].replace('.dat','')+'.h5'))
        self.write('# -------------------------- o --------------------------\n')
        self.report_header(header)

        self.tophit_count = 0

    def report_coarse_channel(self, header,total_n_candi):
        """This function does nothing due to the first line, which returns before the rest of the code can run.
        It was supposed to write info to file if coarse channel option was entered in the commandline
        arguments, but is no longer in use. According to previous documentation, it is supposed to:
        Write header information per given obs.

        Args:
          header: dict
          total_n_candi: int

        Returns:
          

        """

        return None
        
        self.write('# Coarse Channel Number: %i \n' % header['coarse_chan'])
        info_str = '# Number of hits: %i \n'%total_n_candi
        self.write(info_str)

    def report_header(self, header):
        """Write header information per given obs.

        Args:
          header: dict,       information to be written to file header

        Returns:
          : void

        """

        info_str = '# Source:%s\n# MJD: %18.12f\tRA: %s\tDEC: %s\n# DELTAT: %10.6f\tDELTAF(Hz): %10.6f\n'%\
                   (header['SOURCE'],header['MJD'], header['RA'], header['DEC'], header['DELTAT'], header['DELTAF']*1e6)

        self.write(info_str)
        self.write('# --------------------------\n')
        info_str = '# Top_Hit_# \t'
        info_str += 'Drift_Rate \t'
        info_str += 'SNR \t'
        info_str += 'Uncorrected_Frequency \t'
        info_str += 'Corrected_Frequency \t'
        info_str += 'Index \t'
        info_str += 'freq_start \t'
        info_str += 'freq_end \t'
        info_str += 'SEFD \t'
        info_str += 'SEFD_freq \t'
        info_str += 'Coarse_Channel_Number \t'
        info_str += 'Full_number_of_hits \t'
        info_str +='\n'
        self.write(info_str)
        self.write('# --------------------------\n')

    def report_tophit(self, max_val, ind, ind_tuple, tdwidth, fftlen, header,total_n_candi,spec_slice=None,obs_info=None):
        """This function looks into the top hit in a region, basically finds the local maximum and saves that.

        Args:
          max_val: findopp.max_vals,
          ind: int,                    index at which top hit is located in max_val's maxdrift and
        maxsnr
          ind_tuple: tuple(int, int)         (lbound, ubound)
          tdwidth: int,
          fftlen: int,                    length of the fast fourier transform matrix
          header: dict,                   contains info on coarse channel to be written to file
          total_n_candi: int,
          spec_slice: dict, (Default value = None)
          obs_info: dict,                   used to hold info found on file, including info about pulsars,
        RFI, and SEFD (Default value = None)

        Returns:
          : FileWriter object that called this function.

        """

        offset = int((tdwidth - fftlen)/2)
        tdwidth =  len(max_val.maxsnr)

        self.tophit_count += 1
        freq_start = chan_freq(header, ind_tuple[0]-offset, tdwidth, 0)
        freq_end   = chan_freq(header, ind_tuple[1]-1-offset, tdwidth, 0)

        uncorr_freq = chan_freq(header, ind-offset, tdwidth, 0)
        corr_freq = chan_freq(header, ind-offset, tdwidth, 1)

        #Choosing the index of given SEFD and freq.
        if obs_info['SEFDs_freq'][0] > 0.:
            this_one = np.arange(len(obs_info['SEFDs_freq']))[ (obs_info['SEFDs_freq_up']>uncorr_freq) ][0]
        else:
            this_one = 0

        info_str = '%03d\t'%(self.tophit_count)  #Top Hit number
        info_str += '%10.6f\t'%max_val.maxdrift[ind]  #Drift Rate
        info_str += '%10.6f\t'%max_val.maxsnr[ind]  #SNR
        info_str += '%14.6f\t'%uncorr_freq #Uncorrected Frequency:
        info_str += '%14.6f\t'%corr_freq #Corrected Frequency:
        info_str += '%d\t'%(ind - offset) #Index:
        info_str += '%14.6f\t'%freq_start #freq_start:
        info_str += '%14.6f\t'%freq_end #freq_end:
        info_str += '%s\t'%obs_info['SEFDs_val'][this_one] #SEFD:
        info_str += '%14.6f\t'%obs_info['SEFDs_freq'][this_one] #SEFD_mid_freq:
        info_str += '%i\t'%header['coarse_chan']
        info_str += '%i\t'%total_n_candi #
        info_str +='\n'
        self.write(info_str)

        return self

class LogWriter(GeneralWriter):
    """Used to write data to log."""
    def report_candidate(self, info_str):
        """Does nothing. Unimplemented and unused.

        Args:
          info_str: 

        Returns:

        """
        return None

    def info(self, info_str):
        """Writes info_str to file.

        Args:
          info_str: string,     to be written to file

        Returns:
          : void

        """
        self.write(info_str + '\n')
        return None






# To increment version
# Check you have ~/.pypirc filled in
# git tag x.y.z
# git push && git push --tags
# rm -rf dist; python setup.py sdist bdist_wheel
# auditwheel repair dist/*.whl -w dist/ (Linux)
# TEST: twine upload --repository-url https://test.pypi.org/legacy/ dist/*
# twine upload dist/*

from setuptools import setup, find_packages
from Cython.Build import cythonize
from Cython.Distutils import build_ext
import numpy
from setuptools.extension import Extension

__version__ = "1.2.2"

with open("README.md", "r") as fh:
    long_description = fh.read()

with open("requirements.txt", "r") as fh:
    install_requires = fh.readlines()

with open("requirements_test.txt", "r") as fh:
    test_requirements = fh.readlines()

entry_points = {
    'console_scripts' :
        ['turboSETI = turbo_seti.find_doppler.seti_event:main',
         'find_event = turbo_seti.find_event.find_event:main',
         'find_scan_sets = turbo_seti.find_event.find_scan_sets:main',
         'plot_event = turbo_seti.find_event.plot_event:main',
     ]
}

extensions = [Extension(
        name="turbo_seti.find_doppler.taylor_tree",
        sources=["turbo_seti/find_doppler/taylor_tree.pyx"],
        include_dirs=[numpy.get_include()],
        )
    ]
cmdclass = {'build_ext': build_ext}


# Need to copy over index files, generate filenames
idxs = [2,3,4,5,6,7,8,9,10,11]
drift_idxs = ['drift_indexes/drift_indexes_array_%i.txt' % ii for ii in idxs]
package_data={
    'turbo_seti': drift_idxs,
}

setup(
    name="turbo_seti",
    version=__version__,
    packages=find_packages(),
    package_data=package_data,
    include_package_data=True,
    cmdclass=cmdclass,
    ext_modules=cythonize(extensions),
    install_requires=install_requires,
    tests_require=test_requirements,
    entry_points=entry_points,
    author="Emilio Enriquez",
    author_email="e.enriquez@berkeley.edu",
    description="Analysis tool for the search of narrow band drifting signals in filterbank data",
    long_description=long_description,
    long_description_content_type='text/markdown',
    license="MIT License",
    keywords="astronomy",
    url="https://github.com/UCBerkeleySETI/turbo_seti",
    zip_safe=False,
    options={"bdist_wheel": {"universal": "1"}},
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Science/Research",
        "Programming Language :: Python :: 3",
        "Topic :: Scientific/Engineering",
        ]
)



from .find_doppler import seti_event, FindDoppler, helper_functions
from .find_event import find_event, plot_event, find_scan_sets
from pkg_resources import get_distribution, DistributionNotFound

try:
    __version__ = get_distribution('turbo_seti').version
except DistributionNotFound:
    __version__ = '0.0.0 - please install via pip/setup.py'



#!/usr/bin/env python


import numpy as np
import sys
import os
import logging
logger = logging.getLogger(__name__)
import gc   #Garbage collector.

from .data_handler import DATAHandle
from .file_writers import FileWriter, LogWriter
from .helper_functions import *

#For importing cython code
import pyximport
pyximport.install(setup_args={"include_dirs":np.get_include()}, reload_support=True)

try:
    from . import taylor_tree as tt
except:
    import taylor_tree as tt

#For debugging
#import pdb;# pdb.set_trace()
class max_vals:
    """ """
    def __init__(self):
        self.maxsnr = None
        self.maxdrift = None
        self.maxsmooth = None
        self.maxid = None
        self.total_n_hits = None

class hist_vals:
    """Temporary class that saved the normalized spectrum for all drift rates."""
    def __init__(self):
        self.histsnr = None
        self.histdrift = None
        self.histid = None

class FindDoppler:
    """ """
    def __init__(self, datafile, max_drift, min_drift=0, snr=25.0, out_dir='./', coarse_chans=None, obs_info=None, flagging=None, n_coarse_chan=None):
        """
        Initializes FindDoppler object

        Args:
            datafile (string):  Inputted filename (.h5 or .fil)
            max_drift (float):  Max drift rate in Hz/second
            min_drift (int):    Min drift rate in Hz/second
            snr (float):        Signal to Noise Ratio - A ratio bigger than 1 to 1 has more signal than  noise
            out_dir (string):   Directory where output files should be placed. By default this is the
                                current working directory.
            coarse_chans (list(int)):  the inputted comma separated list of coarse channels to analyze, if any.
            obs_info (dict):     Used to hold info found on file, including info about pulsars, RFI, and SEFD
            flagging (bool):     Flags the edges of the PFF for BL data (with 3Hz res per channel)
            n_coarse_chan (int): Number of coarse channels in file
        """
        self.min_drift = min_drift
        self.max_drift = max_drift
        self.snr = snr
        self.out_dir = out_dir

        self.data_handle = DATAHandle(datafile, out_dir=out_dir, n_coarse_chan=n_coarse_chan, coarse_chans=coarse_chans)
        if (self.data_handle is None) or (self.data_handle.status is False):
            raise IOError("File error, aborting...")

        logger.info(self.data_handle.get_info())
        logger.info("A new FinDoppler instance created!")

        if obs_info is None:
            obs_info = {}
            obs_info['pulsar']        = 0  # Bool if pulsar detection.
            obs_info['pulsar_found']  = 0  # Bool if pulsar detection.
            obs_info['pulsar_dm']     = 0.0  # Pulsar expected DM.
            obs_info['pulsar_snr']    = 0.0  # Signal toNoise Ratio (SNR)
            obs_info['pulsar_stats']  = np.zeros(6)
            obs_info['RFI_level']     = 0.0  # Radio Frequency Interference
            obs_info['Mean_SEFD']     = 0.0  # Mean System Equivalent Flux Density
            obs_info['psrflux_Sens']  = 0.0
            obs_info['SEFDs_val']     = [0.0]  # System Equivalent Flux Density values
            obs_info['SEFDs_freq']    = [0.0]  # System Equivalent Flux Density frequency
            obs_info['SEFDs_freq_up'] = [0.0]

        self.obs_info = obs_info

        self.status = True
        self.flagging = flagging

    def get_info(self):
        """Generate info string

        Args:

        Returns:
          : String that contains the values of this FinDoppler object's attributes.

        """
        info_str = "File: %s\n drift rates (min, max): (%f, %f)\n SNR: %f\n"%(self.data_handle.filename, self.min_drift, self.max_drift,self.snr)
        return info_str

    def search(self):
        """Top level search routine"""
        logger.debug("Start searching...")
        logger.debug(self.get_info())

        self.logwriter = LogWriter('%s/%s.log'%(self.out_dir.rstrip('/'),
                                    self.data_handle.data_list[0].filename.split('/')[-1].replace('.h5','').replace('.fits','').replace('.fil','')))
        self.filewriter = FileWriter('%s/%s.dat'%(self.out_dir.rstrip('/'),
                                     self.data_handle.data_list[0].filename.split('/')[-1].replace('.h5','').replace('.fits','').replace('.fil','')),
                                     self.data_handle.data_list[0].header)

        logger.info("Start ET search for %s"%self.data_handle.data_list[0].filename)
        self.logwriter.info("Start ET search for %s"%(self.data_handle.data_list[0].filename))

        for ii,target_data_obj in enumerate(self.data_handle.data_list):
            self.search_data(target_data_obj)
            self.data_handle.data_list[ii].close()
            gc.collect()

    def search_data(self, data_obj):
        """Search the waterfall data of a data handler (coarse channel).

        Args:
          data_obj(DATAH5): File's waterfall data handler

        Returns:

        """
        logger.info("Start searching for coarse channel: %s"%data_obj.header['coarse_chan'])
        self.logwriter.info("Start searching for %s ; coarse channel: %i "%(data_obj.filename,data_obj.header['coarse_chan']))
        spectra, drift_indices = data_obj.load_data()
        tsteps = data_obj.tsteps
        tsteps_valid = data_obj.tsteps_valid
        tdwidth = data_obj.tdwidth
        fftlen = data_obj.fftlen
        nframes = tsteps_valid
        shoulder_size = data_obj.shoulder_size

        if self.flagging:
            ##EE This flags the edges of the PFF for BL data (with 3Hz res per channel).
            ##EE The PFF flat profile falls after around 100k channels.
            ##EE But it falls slowly enough that could use 50-80k channels.
            median_flag = np.median(spectra)
#             spectra[:,:80000] = median_flag/float(tsteps)
#             spectra[:,-80000:] = median_flag/float(tsteps)

            ##EE Flagging spikes in time series.
            time_series=spectra.sum(axis=1)
            time_series_median = np.median(time_series)
            mask=(time_series-time_series_median)/time_series.std() > 10   #Flagging spikes > 10 in SNR

            if mask.any():
                self.logwriter.info("Found spikes in the time series. Removing ...")
                spectra[mask,:] = time_series_median/float(fftlen)  # So that the value is not the median in the time_series.

        else:
            median_flag = np.array([0])

        # allocate array for findopplering
        # init findopplering array to zero
        tree_findoppler = np.zeros(tsteps * tdwidth,dtype=np.float64) + median_flag

        # allocate array for holding original
        # Allocates array in a fast way (without initialize)
        tree_findoppler_original = np.empty_like(tree_findoppler)

        # allocate array for negative doppler rates
        tree_findoppler_flip = np.empty_like(tree_findoppler)

        # build index mask for in-place tree doppler correction
        ibrev = np.zeros(tsteps, dtype=np.int32)

        for i in range(0, tsteps):
            ibrev[i] = bitrev(i, int(np.log2(tsteps)))

##EE: should double check if tdwidth is really better than fftlen here.
        max_val = max_vals()
        if max_val.maxsnr == None:
            max_val.maxsnr = np.zeros(tdwidth, dtype=np.float64)
        if max_val.maxdrift == None:
            max_val.maxdrift = np.zeros(tdwidth, dtype=np.float64)
        if max_val.maxsmooth == None:
            max_val.maxsmooth = np.zeros(tdwidth, dtype='uint8')
        if max_val.maxid == None:
            max_val.maxid = np.zeros(tdwidth, dtype='uint32')
        if max_val.total_n_hits == None:
            max_val.total_n_hits = 0

        #EE: Making "shoulders" to avoid "edge effects". Could do further testing.
        specstart = int(tsteps*shoulder_size/2)
        specend = tdwidth - (tsteps * shoulder_size)

        #--------------------------------
        #Stats calc
        self.the_mean_val, self.the_stddev = comp_stats(spectra.sum(axis=0))

        #--------------------------------
        #Looping over drift_rate_nblock
        #--------------------------------
        drift_rate_nblock = int(np.floor(self.max_drift / (data_obj.drift_rate_resolution*tsteps_valid)))

##EE-debuging        kk = 0

        for drift_block in range(-1*drift_rate_nblock,drift_rate_nblock+1):
            logger.debug( "Drift_block %i"%drift_block)

            #----------------------------------------------------------------------
            # Negative drift rates search.
            #----------------------------------------------------------------------
            if drift_block <= 0:

                #Populates the find_doppler tree with the spectra
                populate_tree(spectra,tree_findoppler,nframes,tdwidth,tsteps,fftlen,shoulder_size,roll=drift_block,reverse=1)

                # populate original array
                np.copyto(tree_findoppler_original, tree_findoppler)

                # populate neg doppler array
                np.copyto(tree_findoppler_flip, tree_findoppler_original)
                
                # Flip matrix across X dimension to search negative doppler drift rates
                FlipX(tree_findoppler_flip, tdwidth, tsteps)
                logger.info("Doppler correcting reverse...")
                tt.taylor_flt(tree_findoppler_flip, tsteps * tdwidth, tsteps)
                logger.debug( "done...")
                
                complete_drift_range = data_obj.drift_rate_resolution*np.array(range(-1*tsteps_valid*(np.abs(drift_block)+1)+1,-1*tsteps_valid*(np.abs(drift_block))+1))
                for k,drift_rate in enumerate(complete_drift_range[(complete_drift_range<self.min_drift) & (complete_drift_range>=-1*self.max_drift)]):
                    # indx  = ibrev[drift_indices[::-1][k]] * tdwidth

                    # DCP 2020.04 -- WAR to drift rate in flipped files
                    if data_obj.header['DELTAF'] < 0:
                        drift_rate *= -1

                    indx  = ibrev[drift_indices[::-1][(complete_drift_range<self.min_drift) & (complete_drift_range>=-1*self.max_drift)][k]] * tdwidth

                    # SEARCH NEGATIVE DRIFT RATES
                    spectrum = tree_findoppler_flip[indx: indx + tdwidth]

                    # normalize
                    spectrum -= self.the_mean_val
                    spectrum /= self.the_stddev

                    #Reverse spectrum back
                    spectrum = spectrum[::-1]

                    n_hits, max_val = hitsearch(spectrum, specstart, specend, self.snr, drift_rate, data_obj.header, fftlen, tdwidth, max_val, 0)
                    info_str = "Found %d hits at drift rate %15.15f\n"%(n_hits, drift_rate)
                    max_val.total_n_hits += n_hits
                    logger.debug(info_str)
                    self.logwriter.info(info_str)

            #----------------------------------------------------------------------
            # Positive drift rates search.
            #----------------------------------------------------------------------
            if drift_block >= 0:

                #Populates the find_doppler tree with the spectra
                populate_tree(spectra,tree_findoppler,nframes,tdwidth,tsteps,fftlen,shoulder_size,
                              roll=drift_block,reverse=1)

                # populate original array
                np.copyto(tree_findoppler_original, tree_findoppler)

                logger.info("Doppler correcting forward...")
                tt.taylor_flt(tree_findoppler, tsteps * tdwidth, tsteps)
                logger.debug( "done...")
                if (tree_findoppler == tree_findoppler_original).all():
                     logger.error("taylor_flt has no effect?")
                else:
                     logger.debug("tree_findoppler changed")

                ##EE: Calculates the range of drift rates for a full drift block.
                complete_drift_range = data_obj.drift_rate_resolution*np.array(range(tsteps_valid*(drift_block),tsteps_valid*(drift_block +1)))

                for k,drift_rate in enumerate(complete_drift_range[(complete_drift_range>=self.min_drift) & (complete_drift_range<=self.max_drift)]):

                    indx  = ibrev[drift_indices[k]] * tdwidth

                    #DCP 2020.04 -- WAR to drift rate in flipped files
                    if data_obj.header['DELTAF'] < 0:
                        drift_rate *= -1

                    # SEARCH POSITIVE DRIFT RATES
                    spectrum = tree_findoppler[indx: indx+tdwidth]

                    # normalize
                    spectrum -= self.the_mean_val
                    spectrum /= self.the_stddev

                    n_hits, max_val = hitsearch(spectrum, specstart, specend, self.snr, drift_rate, data_obj.header, fftlen, tdwidth, max_val, 0)
                    info_str = "Found %d hits at drift rate %15.15f\n"%(n_hits, drift_rate)
                    max_val.total_n_hits += n_hits
                    logger.debug(info_str)
                    self.logwriter.info(info_str)

        # Writing the top hits to file.
        self.filewriter = tophitsearch(tree_findoppler_original, max_val, tsteps, nframes, data_obj.header, tdwidth,
                                       fftlen, self.max_drift,data_obj.obs_length, out_dir = self.out_dir,
                                       logwriter=self.logwriter, filewriter=self.filewriter, obs_info=self.obs_info)

        logger.info("Total number of candidates for coarse channel "+ str(data_obj.header['coarse_chan']) +" is: %i"%max_val.total_n_hits)

#  ======================================================================  #

def populate_tree(spectra,tree_findoppler,nframes,tdwidth,tsteps,fftlen,shoulder_size,roll=0,reverse=0):
    """This script populates the findoppler tree with the spectra.
    It creates two "shoulders" (each region of tsteps*(shoulder_size/2) in size) to avoid "edge" issues.
    It uses np.roll() for drift-rate blocks higher than 1.

    Args:
      spectra: ndarray,        spectra calculated from file
      tree_findoppler: ndarray,        tree to be populated with spectra
      nframes: int,
      tdwidth: int,
      tsteps: int,
      fftlen: int,            length of fast fourier transform (fft) matrix
      shoulder_size: int,            size of shoulder region
      roll: int,            used to calculate amount each entry to the spectra should be rolled (shifted) (Default value = 0)
      reverse: int(boolean),   used to determine which way spectra should be rolled (shifted) (Default value = 0)

    Returns:
      : ndarray,        spectra-populated version of the input tree_findoppler

    """

    if reverse:
        direction = -1
    else:
        direction = 1

    for i in range(0, nframes):
        sind = tdwidth*i + tsteps*int(shoulder_size/2)
        cplen = fftlen

        ##EE copy spectra into tree_findoppler, leaving two regions in each side blanck (each region of tsteps*(shoulder_size/2) in size).
        # Copy spectra into tree_findoppler, with rolling.
        np.copyto(tree_findoppler[sind: sind + cplen], np.roll(spectra[i],roll*i*direction))

#         ##EE loads the end part of the current spectrum into the left hand side black region in tree_findoppler (comment below says "next spectra" but for that need i+1...bug?)
         #load end of current spectra into left hand side of next spectra
        sind = i * tdwidth
        np.copyto(tree_findoppler[sind: sind + tsteps*int(shoulder_size/2)], spectra[i, fftlen-(tsteps*int(shoulder_size/2)):fftlen])

    return tree_findoppler



def hitsearch(spectrum, specstart, specend, hitthresh, drift_rate, header, fftlen, tdwidth, max_val, reverse):
    """Searches for hits at given drift rate. A hit occurs in each channel if > hitthresh.

    Args:
      spectrum: ndarray,
      specstart: int,                first index to search for hit in spectrum
      specend: int,                last index to search for hit in spectrum
      hitthresh: float,              signal to noise ratio used as threshold for determining hits
      drift_rate: float,              drift rate at which we are searching for hits
      header: dict,               header in fits header format. See data_handler.py's DATAH5 class header
      fftlen: int,                UNUSED
      tdwidth: int,
      max_val: max_vals,           object to be filled with max values from this search and then returned
      reverse: int(boolean),       used to flag whether fine channel should be reversed

    Returns:
      : int, max_vals,      j is the amount of hits.

    """

    logger.debug('Start searching for hits at drift rate: %f'%drift_rate)
    j = 0
    for i in (spectrum[specstart:specend] > hitthresh).nonzero()[0] + specstart:
        k =  (tdwidth - 1 - i) if reverse else i
        info_str = 'Hit found at SNR %f! %s\t'%(spectrum[i], '(reverse)' if reverse else '')
        info_str += 'Spectrum index: %d, Drift rate: %f\t'%(i, drift_rate)
        info_str += 'Uncorrected frequency: %f\t'%chan_freq(header,  k, tdwidth, 0)
        info_str += 'Corrected frequency: %f'%chan_freq(header, k, tdwidth, 1)
        logger.debug(info_str)
        j += 1
        used_id = j
        if spectrum[i] > max_val.maxsnr[k]:
            max_val.maxsnr[k] = spectrum[i]
            max_val.maxdrift[k] = drift_rate
            max_val.maxid[k] = used_id

    return j, max_val

def tophitsearch(tree_findoppler_original, max_val, tsteps, nframes, header, tdwidth, fftlen,max_drift,obs_length, out_dir='', logwriter=None, filewriter=None,obs_info=None):
    """This finds the hits with largest SNR within 2*tsteps frequency channels.

    Args:
      tree_findoppler_original: ndarray,        spectra-populated findoppler tree
      max_val: max_vals,       contains max values from hitsearch
      tsteps: int,
      nframes: int,            UNUSED
      header: dict,           header in fits header format. See data_handler.py's DATAH5 class header. Used to report tophit in filewriter
      tdwidth: int,
      fftlen: int,            length of fast fourier transform (fft) matrix
      max_drift: float,          Max drift rate in Hz/second
      obs_length: float,
      out_dir: string,         UNUSED (Default value = '')
      logwriter: LogWriter       logwriter to which we should write if we find a top hit (Default value = None)
      filewriter: FileWriter      filewriter corresponding to file to which we should save the
    local maximum of tophit. See report_tophit in filewriters.py (Default value = None)
      obs_info: dict, (Default value = None)

    Returns:
      : FileWriter,     same filewriter that was input

    """

    maxsnr = max_val.maxsnr
    logger.debug("original matrix size: %d\t(%d, %d)"%(len(tree_findoppler_original), tsteps, tdwidth))
    tree_orig = tree_findoppler_original.reshape((tsteps, tdwidth))
    logger.debug("tree_orig shape: %s"%str(tree_orig.shape))

    for i in (maxsnr > 0).nonzero()[0]:
        lbound = int(max(0, i - obs_length*max_drift/2))
        ubound = int(min(tdwidth, i + obs_length*max_drift/2))

        skip = 0

        if (maxsnr[lbound:ubound] > maxsnr[i]).nonzero()[0].any():
            skip = 1

        if skip:
            logger.debug("SNR not big enough... %f pass... index: %d"%(maxsnr[i], i))
        else:
            info_str = "Top hit found! SNR: %f ... index: %d"%(maxsnr[i], i)
            logger.info(info_str)
            if logwriter:
                logwriter.info(info_str)
                #logwriter.report_tophit(max_val, i, header)
#EE            logger.debug("slice of spectrum...size: %s"%str(tree_orig[0:nframes, lbound:ubound].shape))
            if filewriter:
                filewriter = filewriter.report_tophit(max_val, i, (lbound, ubound), tdwidth, fftlen, header,max_val.total_n_hits,obs_info=obs_info)
#EE: not passing array cut, since not saving in .dat file                filewriter = filewriter.report_tophit(max_val, i, (lbound, ubound), tree_orig[0:nframes, lbound:ubound], header)

##EE : Uncomment if want to save each blob              np.save(out_dir + '/spec_drift_%.4f_id_%d.npy'%(max_val.maxdrift[i],i), tree_orig[0:nframes, lbound:ubound])
            else:
                logger.error('Not have filewriter? tell me why.')

    return filewriter
